from pyspedas.utilities.dailynames import dailynames
from pyspedas.utilities.download import download
from pyspedas.tplot_tools import time_clip as tclip
from pyspedas.tplot_tools import cdf_to_tplot
from typing import List, Union
from .config import CONFIG

def load(trange:List[str]=['2018-11-5', '2018-11-6'],
         probe:Union[str,List[str]]='1',
         instrument:str='fgm',
         datatype:str='up',
         prefix:str = '',
         suffix:str='',
         get_support_data:bool=False,
         varformat:str=None,
         varnames:List[str]=[],
         downloadonly:bool=False,
         notplot:bool=False,
         no_update:bool=False,
         time_clip:bool=False,
         force_download=False) -> List[str]:
    """
    Load data from Cluster instruments

    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format
            ['YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']
            Default: ['2018-11-5', '2018-11-6']

        probe: list of str
            List of probes to load.  Valid options: '1','2','3','4'
            Default: '1'

        instrument: str
            Instrument to load.
            Default: 'fgm;

        datatype: str
            Data type; Valid options:
            Default: 'up'

        prefix: str
            The tplot variable names will be given this prefix.
            Default: ''

        suffix: str
            The tplot variable names will be given this suffix.
            Default: ''

        get_support_data: bool
            If True, Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted. If empty or None, all variables will be loaded.
            Default: None (all variables loaded)

        varnames: list of str
            List of CDF variable names to load (if empty or not specified,
            all data variables are loaded)
            Default: [] (all variables loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into
            tplot variables
            Default: False

        notplot: bool
            Return the data in hash tables instead of creating tplot variables
            Default: False

        no_update: bool
            If set, only load data from your local cache
            Default: False

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword
            Default: False

        force_download: bool
            Download file even if local version is more recent than server version
            Default: False


    Returns
    -------
        list of str
            List of tplot variables created.

    Notes
    -----
    This function loads data from the Cluster mission; this function is not meant 
    to be called directly; instead, see the wrappers:
        pyspedas.projects.cluster.fgm
        pyspedas.projects.cluster.aspoc
        pyspedas.projects.cluster.cis
        pyspedas.projects.cluster.dwp
        pyspedas.projects.cluster.edi
        pyspedas.projects.cluster.efw
        pyspedas.projects.cluster.peace
        pyspedas.projects.cluster.rapid
        pyspedas.projects.cluster.staff
        pyspedas.projects.cluster.wbd
        pyspedas.projects.cluster.whi

    """
    if prefix is None:
        prefix = ''
    if suffix is None:
        suffix = ''

    if not isinstance(probe, list):
        probe = [probe]

    probe = [str(prb) for prb in probe] # these will need to be strings from now on

    out_files = []

    res=24*3600

    if instrument != 'wbd':
        # note: can't use last_version with WBD data due to using wild cards for the times (and not just in the version)
        last_version = True
    else:
        last_version = False

    for prb in probe:
        if instrument == 'fgm':
            if datatype == 'cp':
                pathformat = 'c'+prb+'/cp/%Y/c'+prb+'_cp_fgm_spin_%Y%m%d_v??.cdf'
            else:
                pathformat = 'c'+prb+'/'+datatype+'/'+instrument+'/%Y/c'+prb+'_'+datatype+'_'+instrument+'_%Y%m%d_v??.cdf'
        elif instrument == 'aspoc':
            pathformat = 'c'+prb+'/'+datatype+'/asp/%Y/c'+prb+'_'+datatype+'_asp_%Y%m%d_v??.cdf'
        elif instrument == 'cis':
            pathformat = 'c'+prb+'/'+datatype+'/'+instrument+'/%Y/c'+prb+'_'+datatype+'_'+instrument+'_%Y%m%d_v??.cdf'
        elif instrument == 'dwp':
            pathformat = 'c'+prb+'/'+datatype+'/'+instrument+'/%Y/c'+prb+'_'+datatype+'_'+instrument+'_%Y%m%d_v??.cdf'
        elif instrument == 'edi':
            pathformat = 'c'+prb+'/'+datatype+'/'+instrument+'/%Y/c'+prb+'_'+datatype+'_'+instrument+'_%Y%m%d_v??.cdf'
        elif instrument == 'efw':
            pathformat = 'c'+prb+'/'+datatype+'/'+instrument+'/%Y/c'+prb+'_'+datatype+'_'+instrument+'_%Y%m%d_v??.cdf'
        elif instrument == 'peace':
            pathformat = 'c'+prb+'/'+datatype+'/pea/%Y/c'+prb+'_'+datatype+'_pea_%Y%m%d_v??.cdf'
        elif instrument == 'rapid':
            pathformat = 'c'+prb+'/'+datatype+'/rap/%Y/c'+prb+'_'+datatype+'_rap_%Y%m%d_v??.cdf'
        elif instrument == 'staff':
            pathformat = 'c'+prb+'/'+datatype+'/sta/%Y/c'+prb+'_'+datatype+'_sta_%Y%m%d_v??.cdf'
        elif instrument == 'whi':
            pathformat = 'c'+prb+'/'+datatype+'/'+instrument+'/%Y/c'+prb+'_'+datatype+'_'+instrument+'_%Y%m%d_v??.cdf'
        elif instrument == 'wbd':
            pathformat = 'c'+prb+'/'+instrument+'/%Y/%m/c'+prb+'_'+datatype+'_'+instrument+'_%Y%m%d%H%M_v??.cdf'
            res = 600.0

        # find the full remote path names using the trange
        remote_names = dailynames(file_format=pathformat, trange=trange, res=res)

        files = download(remote_file=remote_names, remote_path=CONFIG['remote_data_dir'], local_path=CONFIG['local_data_dir'], no_download=no_update, last_version=last_version, force_download=force_download)
        if files is not None:
            for file in files:
                out_files.append(file)

    out_files = sorted(out_files)

    if downloadonly:
        return out_files

    tvars = cdf_to_tplot(out_files, prefix=prefix, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, notplot=notplot)

    if notplot:
        return tvars

    if time_clip:
        for new_var in tvars:
            tclip(new_var, trange[0], trange[1], suffix='')

    return tvars



import unittest
import pyspedas
from pyspedas.tplot_tools import data_exists, tplot_rename, set_coords
from pyspedas.projects.mms.cotrans.mms_qcotrans import mms_qcotrans
from pyspedas.projects.mms.cotrans.mms_cotrans_lmn import mms_cotrans_lmn


class CotransTestCases(unittest.TestCase):
    def test_qcotrans_sm_to_gse(self):
        pyspedas.projects.mms.mec()
        ret1 = mms_qcotrans('mms1_mec_v_sm', 'mms1_mec_v_sm_2gse', out_coord='gse')
        ret2 = mms_qcotrans('mms1_mec_r_sm', 'mms1_mec_r_sm_2gse', out_coord='gse')
        self.assertTrue(data_exists('mms1_mec_v_sm_2gse'))
        self.assertTrue('mms1_mec_v_sm_2gse' in ret1)
        self.assertTrue('mms1_mec_r_sm_2gse' in ret2)
        self.assertTrue(data_exists('mms1_mec_r_sm_2gse'))
        mms_qcotrans(['mms1_mec_r_sm', 'mms1_mec_v_sm'], ['mms1_mec_r_sm_2gse', 'mms1_mec_v_sm_2gse'], out_coord=['gse', 'gse'])

    def test_qcotrans_fgm_sm_to_gse(self):
        pyspedas.projects.mms.mec()
        pyspedas.projects.mms.fgm()
        mms_qcotrans('mms1_fgm_b_gsm_srvy_l2_bvec', 'mms1_fgm_b_gsm_brst_l2_bvec_2gse', probe=1, out_coord='gse')
        self.assertTrue(data_exists('mms1_fgm_b_gsm_brst_l2_bvec_2gse'))

    def test_qcotrans_errors(self):
        with self.assertLogs(level='WARNING') as log:
            pyspedas.projects.mms.mec()

            # in_name not specified
            mms_qcotrans(out_name='mms1_mec_v_sm_2gse', out_coord='gse')
            self.assertIn("Input variable name is missing", log.output[0])

            # out_name not specified
            mms_qcotrans(in_name='mms1_mec_v_sm_2gse', out_coord='gse')
            self.assertIn("Output variable name is missing", log.output[1])

            # in_coord not specified, and not set in metadata
            set_coords('mms1_mec_v_sm', '')
            mms_qcotrans(in_name='mms1_mec_v_sm', out_name='mms1_mec_v_sm_2gse', out_coord='gse')
            self.assertIn("Could not determine coordinate system for: mms1_mec_v_sm", log.output[2])

            # invalid in_coord
            set_coords('mms1_mec_v_sm', '')
            mms_qcotrans(in_name='mms1_mec_v_sm', out_name='mms1_mec_v_sm_2gse2', out_coord='gse')
            self.assertIn("Could not determine coordinate system for: mms1_mec_v_sm", log.output[3])

            # invalid out_coord
            set_coords('mms1_mec_v_sm', 'sm')
            mms_qcotrans(in_name='mms1_mec_v_sm', out_name='mms1_mec_v_sm_2gse2', out_coord='gse2')
            self.assertIn("Unsupported output coordinate system: gse2", log.output[4])

            # trouble extracting probe from var name
            tplot_rename('mms1_mec_v_sm', 'mmsx_mec_v_sm')
            mms_qcotrans(in_name='mmsx_mec_v_sm', out_name='mms1_mec_v_sm_2gse2', out_coord='gse2')
            tplot_rename('mms1_mec_v_sm', 'smvar')
            mms_qcotrans(in_name='smvar', out_name='mms1_mec_v_sm_2gse', out_coord='gse')
            self.assertIn("Problem occurred during the transformation", log.output[5])
            self.assertIn("Unsupported output coordinate system: gse2", log.output[6])
            self.assertIn("Unknown probe for variable: mmsx_mec_v_sm", log.output[7])
            self.assertIn("Could not determine coordinate system for: smvar", log.output[8])

            # should warn when you're transforming to ssl/bcs coordinates
            mms_qcotrans(out_name='mms1_mec_v_sm_2ssl', out_coord='ssl')
            mms_qcotrans(out_name='mms1_mec_v_sm_2bcs', out_coord='bcs')
            self.assertIn("Input variable name is missing", log.output[9])
            self.assertIn("Input variable name is missing", log.output[10])

            # unsupported coordinate system
            mms_qcotrans(in_name='mms1_mec_v_sm', out_name='mms1_mec_v_sm_2gse', in_coord='unsupported', out_coord='gse')
            self.assertIn("Unsupported input coordinate system: unsupported", log.output[11])
            


    def test_lmn(self):
        pyspedas.projects.mms.fgm(trange=['2015-10-16/13:00', '2015-10-16/13:10'], data_rate='brst')
        mms_cotrans_lmn('mms1_fgm_b_gsm_brst_l2_bvec', 'mms1_fgm_b_gsm_brst_l2_bvec_2lmn')
        self.assertTrue(data_exists('mms1_fgm_b_gsm_brst_l2_bvec_2lmn'))

    def test_lmn_errors(self):
        with self.assertLogs(level='ERROR') as log:
            pyspedas.projects.mms.fgm(trange=['2015-10-16/13:00', '2015-10-16/13:10'], data_rate='brst')

            # invalid variable name
            mms_cotrans_lmn('mms1_fgm_b_gsm_brst_l2_bvec2', 'mms1_fgm_b_gsm_brst_l2_bvec_2lmn')
            self.assertIn("Error reading tplot variable: mms1_fgm_b_gsm_brst_l2_bvec2", log.output[0])

            # invalid coordinate system
            set_coords('mms1_fgm_b_gsm_brst_l2_bvec', 'gse2')
            mms_cotrans_lmn('mms1_fgm_b_gsm_brst_l2_bvec', 'mms1_fgm_b_gsm_brst_l2_bvec_2lmn')
            self.assertIn("Please specify the coordinate system of the input data", log.output[1])

            # problem extracting probe from variable name
            tplot_rename('mms1_fgm_b_gsm_brst_l2_bvec', 'mmsx_fgm_b_gsm_brst_l2_bvec')
            mms_cotrans_lmn('mmsx_fgm_b_gsm_brst_l2_bvec', 'mms1_fgm_b_gsm_brst_l2_bvec_2lmn')
            self.assertIn("Please specify the coordinate system of the input data", log.output[2])            


if __name__ == '__main__':
    unittest.main()




import unittest
import numpy as np

from ..mms_file_filter import mms_file_filter

TEST_DATA = ['C:\\Users\\admin\\data\\mms\\mms1\\dfg\\srvy\\ql\\2016\\01\\mms1_dfg_srvy_ql_20160115_v2.13.2.cdf',
    'C:\\Users\\admin\\data\\mms\\mms1\\dfg\\srvy\\ql\\2016\\01\\mms1_dfg_srvy_ql_20160116_v2.13.2.cdf',
    'C:\\Users\\admin\\data\\mms\\mms1\\dfg\\srvy\\ql\\2016\\01\\mms1_dfg_srvy_ql_20160117_v2.13.3.cdf',
    'C:\\Users\\admin\\data\\mms\\mms1\\dfg\\srvy\\ql\\2016\\01\\mms1_dfg_srvy_ql_20160118_v2.13.2.cdf',
    'C:\\Users\\admin\\data\\mms\\mms1\\dfg\\srvy\\ql\\2016\\01\\mms1_dfg_srvy_ql_20160119_v2.14.1.cdf',
    'C:\\Users\\admin\\data\\mms\\mms1\\dfg\\srvy\\ql\\2016\\01\\mms1_dfg_srvy_ql_20160120_v2.14.1.cdf']

class FileFilterTestCases(unittest.TestCase):
    def test_version_eq(self):
        version_eqs =  ['C:\\Users\\admin\\data\\mms\\mms1\\dfg\\srvy\\ql\\2016\\01\\mms1_dfg_srvy_ql_20160115_v2.13.2.cdf',
                        'C:\\Users\\admin\\data\\mms\\mms1\\dfg\\srvy\\ql\\2016\\01\\mms1_dfg_srvy_ql_20160116_v2.13.2.cdf',
                        'C:\\Users\\admin\\data\\mms\\mms1\\dfg\\srvy\\ql\\2016\\01\\mms1_dfg_srvy_ql_20160118_v2.13.2.cdf']
        version_eq_files = mms_file_filter(TEST_DATA, version='2.13.2')
        self.assertTrue(np.array_equal(version_eqs, version_eq_files))

    def test_latest_version(self):
        latest_version = ['C:\\Users\\admin\\data\\mms\\mms1\\dfg\\srvy\\ql\\2016\\01\\mms1_dfg_srvy_ql_20160119_v2.14.1.cdf',
                      'C:\\Users\\admin\\data\\mms\\mms1\\dfg\\srvy\\ql\\2016\\01\\mms1_dfg_srvy_ql_20160120_v2.14.1.cdf']
        latest_files = mms_file_filter(TEST_DATA, latest_version=True)
        self.assertTrue(np.array_equal(latest_version, latest_files))

    def test_min_version(self):
        min_v = ['C:\\Users\\admin\\data\\mms\\mms1\\dfg\\srvy\\ql\\2016\\01\\mms1_dfg_srvy_ql_20160117_v2.13.3.cdf',
                'C:\\Users\\admin\\data\\mms\\mms1\\dfg\\srvy\\ql\\2016\\01\\mms1_dfg_srvy_ql_20160119_v2.14.1.cdf',
                'C:\\Users\\admin\\data\\mms\\mms1\\dfg\\srvy\\ql\\2016\\01\\mms1_dfg_srvy_ql_20160120_v2.14.1.cdf']
        min_versions = mms_file_filter(TEST_DATA, min_version='2.13.3')
        self.assertTrue(np.array_equal(min_versions, min_v))

    def test_major_version(self):
        all_versions = ['C:\\Users\\admin\\data\\mms\\mms1\\dfg\\srvy\\ql\\2016\\01\\mms1_dfg_srvy_ql_20160117_v2.13.3.cdf',
             'C:\\Users\\admin\\data\\mms\\mms1\\dfg\\srvy\\ql\\2016\\01\\mms1_dfg_srvy_ql_20160119_v2.14.1.cdf', 
             'C:\\Users\\admin\\data\\mms\\mms1\\dfg\\srvy\\ql\\2016\\01\\mms1_dfg_srvy_ql_20160120_v2.14.1.cdf', 
             'C:\\Users\\admin\\data\\mms\\mms1\\dfg\\srvy\\ql\\2016\\01\\mms1_dfg_srvy_ql_20160120_v1.14.1.cdf', 
             'C:\\Users\\admin\\data\\mms\\mms1\\dfg\\srvy\\ql\\2016\\01\\mms1_dfg_srvy_ql_20160120_v1.14.1.cdf', 
             'C:\\Users\\admin\\data\\mms\\mms1\\dfg\\srvy\\ql\\2016\\01\\mms1_dfg_srvy_ql_20160120_v1.14.1.cdf']
        major_versions_real = ['C:\\Users\\admin\\data\\mms\\mms1\\dfg\\srvy\\ql\\2016\\01\\mms1_dfg_srvy_ql_20160117_v2.13.3.cdf',
             'C:\\Users\\admin\\data\\mms\\mms1\\dfg\\srvy\\ql\\2016\\01\\mms1_dfg_srvy_ql_20160119_v2.14.1.cdf', 
             'C:\\Users\\admin\\data\\mms\\mms1\\dfg\\srvy\\ql\\2016\\01\\mms1_dfg_srvy_ql_20160120_v2.14.1.cdf']

        filtered_files = mms_file_filter(all_versions, major_version=True)
        self.assertTrue(np.array_equal(filtered_files, major_versions_real))

if __name__ == '__main__':
    unittest.main()


import logging
import numpy as np
def xdegap(var):
    """
    Replace NaN values in an array.

    Parameters
    ----------
    var : numpy.ndarray
        Input array that may contain NaN values.

    Returns
    -------
    numpy.ndarray
        Array with NaN values replaced by the average of surrounding non-NaN values.

    """
    if np.all(np.isnan(var)):
        logging.info("All values are NaN. No replacement performed.")
        return var

    # Iterate through the array and replace NaN values
    for i in range(len(var)):
        if np.isnan(var[i]):
            # Find the first non-NaN value before the current index
            before = np.where(~np.isnan(var[:i]))[0]
            # Find the first non-NaN value after the current index
            after = np.where(~np.isnan(var[i+1:]))[0]

            if len(before) > 0 and len(after) > 0:
                # Both before and after values exist
                before_value = var[before[-1]]
                after_value = var[i + 1 + after[0]]
                var[i] = (before_value + after_value) / 2.0
            elif len(before) > 0:
                # Only before value exists
                var[i] = var[before[-1]]
            elif len(after) > 0:
                # Only after value exists
                var[i] = var[i + 1 + after[0]]

    return var



import os

CONFIG = {'local_data_dir': 'dscovr_data/',
          'remote_data_dir': 'https://spdf.gsfc.nasa.gov/pub/data/dscovr/'}

# override local data directory with environment variables
if os.environ.get('DSC_DATA_DIR'):
    CONFIG['local_data_dir'] = os.environ['DSC_DATA_DIR']
elif os.environ.get('SPEDAS_DATA_DIR'):
    CONFIG['local_data_dir'] = os.sep.join([os.environ['SPEDAS_DATA_DIR'], 'dscovr'])




import os
import unittest
import pyspedas
from pyspedas.tplot_tools import del_data, data_exists, time_double, get_data


class LoadTestCases(unittest.TestCase):
    def test_downloadonly(self):
        mag_files = pyspedas.projects.barrel.sspc(probe='1A', downloadonly=True)
        self.assertTrue(os.path.exists(mag_files[0]))

    def test_sspc(self):
        del_data("*")
        sspc_vars = pyspedas.projects.barrel.sspc(probe='1A')
        self.assertTrue(len(sspc_vars) > 0)
        self.assertTrue('brl1A_SSPC' in sspc_vars)
        self.assertTrue(data_exists('brl1A_SSPC'))

    def test_sspc_time_clip(self):
        del_data("*")
        trange = ['2013-01-29 09:00', '2013-01-29 12:00']
        sspc_vars = pyspedas.projects.barrel.sspc(probe='1A',trange=trange,time_clip=True)
        self.assertTrue(len(sspc_vars) > 0)
        self.assertTrue('brl1A_SSPC' in sspc_vars)
        self.assertTrue(data_exists('brl1A_SSPC'))
        dat=get_data('brl1A_SSPC')
        trange_double=time_double(trange)
        self.assertTrue(dat.times[0] >= trange_double[0])
        self.assertTrue(dat.times[-1] <= trange_double[1])

    def test_mspc(self):
        del_data("*")
        mspc_vars = pyspedas.projects.barrel.mspc(probe='1A')
        self.assertTrue(len(mspc_vars) > 0)
        self.assertTrue('brl1A_MSPC' in mspc_vars)
        self.assertTrue(data_exists('brl1A_MSPC'))

    def test_fspc(self):
        del_data("*")
        fspc_vars = pyspedas.projects.barrel.fspc(probe='1A')
        self.assertTrue(len(fspc_vars) > 0)
        self.assertTrue('brl1A_FSPC1' in fspc_vars)
        self.assertTrue(data_exists('brl1A_FSPC1'))

    def test_rcnt(self):
        del_data("*")
        rcnt_vars = pyspedas.projects.barrel.rcnt(probe='1A')
        self.assertTrue(len(rcnt_vars) > 0)
        self.assertTrue('brl1A_PeakDet' in rcnt_vars)
        self.assertTrue(data_exists('brl1A_PeakDet'))

    def test_magn(self):
        del_data("*")
        magn_vars = pyspedas.projects.barrel.magn(probe='1A')
        self.assertTrue(len(magn_vars) > 0)
        self.assertTrue('brl1A_MAG_X_uncalibrated' in magn_vars)
        self.assertTrue(data_exists('brl1A_MAG_X_uncalibrated'))

    def test_magn_prefix_none(self):
        del_data("*")
        magn_vars = pyspedas.projects.barrel.magn(probe='1A', prefix=None)
        self.assertTrue(len(magn_vars) > 0)
        self.assertTrue('brl1A_MAG_X_uncalibrated' in magn_vars)
        self.assertTrue(data_exists('brl1A_MAG_X_uncalibrated'))

    def test_magn_suffix_none(self):
        del_data("*")
        magn_vars = pyspedas.projects.barrel.magn(probe='1A', suffix=None)
        self.assertTrue(len(magn_vars) > 0)
        self.assertTrue('brl1A_MAG_X_uncalibrated' in magn_vars)
        self.assertTrue(data_exists('brl1A_MAG_X_uncalibrated'))

    def test_magn_prefix_suffix(self):
        del_data("*")
        magn_vars = pyspedas.projects.barrel.magn(probe='1A', prefix='pre_', suffix='_suf')
        self.assertTrue(len(magn_vars) > 0)
        self.assertTrue('pre_brl1A_MAG_X_uncalibrated_suf' in magn_vars)
        self.assertTrue(data_exists('pre_brl1A_MAG_X_uncalibrated_suf'))

    def test_ephm(self):
        del_data("*")
        ephm_vars = pyspedas.projects.barrel.ephm(probe='1A')
        self.assertTrue('brl1A_GPS_Lat' in ephm_vars)
        self.assertTrue(data_exists('brl1A_GPS_Lat'))

    def test_hkpg(self):
        del_data("*")
        hkpg_vars = pyspedas.projects.barrel.hkpg(probe='1A')
        self.assertTrue(len(hkpg_vars) > 0)
        self.assertTrue('brl1A_V0_VoltAtLoad' in hkpg_vars)
        self.assertTrue(data_exists('brl1A_V0_VoltAtLoad'))


if __name__ == '__main__':
    unittest.main()



# Copyright 2018 Regents of the University of Colorado. All Rights Reserved.
# Released under the MIT license.
# This software was developed at the University of Colorado's Laboratory for Atmospheric and Space Physics.
# Verify current version before use at: https://github.com/MAVENSDC/PyTplot

import os
import pickle
import numpy as np
import pyspedas
from pyspedas.tplot_tools import options, store_data
from pyspedas.tplot_tools.tplot_options import tplot_options
from scipy.io import readsav
import logging


def tplot_restore(filename):
    """
    Restore tplot variables that have been saved to a file.

    If the filename has a suffix ".tplot", it is assumed to be in the IDL .sav format as written by the IDL tplot_save routine.
    In this case, it is read using the scipy.io.readsav() routine.

    For any other filename suffix, the file is assumed to be a Python pickle file, as written by the PySPEDAS tplot_save routine.

    Most of the metadata (units, coordinate systems, CDF attributes etc.) should be correctly represented in the
    tplot variables created by tplot_restore.

    Some IDL plot options may not have direct counterparts in
    the PySPEDAS matplotlib-based plot options.

    For multidimensional variables (spectra, particle distributions, etc.), the IDL v, v1, v2 (etc.) attributes should also be correctly
    represented in the output tplot variables.

    Parameters
    -----------
        filename : str
            The path to the ".tplot" file created by IDL tplot_save, or for any other suffix, the pickle file created by PySPEDAS tplot_save.
            
    Returns
    --------

        None
    
    Examples
    --------
        >>> # Restore the saved data from the tplot_save example
        >>> import pyspedas
        >>> pyspedas.tplot_restore('C:/temp/variable1.pyspedas')

    """
    
    #Error check
    if not (os.path.isfile(filename)):
        logging.error("tplot_restore: %s is not a valid file name",filename)
        return
    
    #Check if the restored file was an IDL file
    
    if filename.endswith('.tplot'):
        temp_tplot = readsav(filename)
        for i in range(len(temp_tplot['dq'])):
            if isinstance(temp_tplot['dq'][i][0], str):
                logging.warning("tplot_restore: Error reading variable; this error occurs when the variable wasn't loaded in IDL when the SAV file was created.")
                continue

            data_name = temp_tplot['dq'][i][0].decode("utf-8")
            try:
                temp_x_data = temp_tplot['dq'][i][1][0][0].squeeze()
            except AttributeError as err:
                logging.warning("tplot_restore: Attribute error squeezing temp_x_data, index %d, name %s. (Value is scalar rather than array?)",i,data_name)
                continue


            #Pandas reads in data the other way I guess
            if len(temp_tplot['dq'][i][1][0][2].shape) == 4:
                temp_y_data = np.transpose(temp_tplot['dq'][i][1][0][2], axes=(3, 2, 1, 0))
            elif len(temp_tplot['dq'][i][1][0][2].shape) == 3:
                temp_y_data = np.transpose(temp_tplot['dq'][i][1][0][2], axes=(2, 1, 0))
            elif len(temp_tplot['dq'][i][1][0][2].shape) == 2:
                temp_y_data = np.transpose(temp_tplot['dq'][i][1][0][2])
            else:
                temp_y_data = temp_tplot['dq'][i][1][0][2]

            # Check for loss of leading dimension if only 1 timestamp is present
            shape = temp_x_data.shape
            if len(shape) == 0:
                temp_x_data = np.atleast_1d(temp_x_data)

            # variable contains V1, V2 and V3 (e.g., DF as a function of energy, theta, phi)
            if len(temp_tplot['dq'][i][1][0]) == 10:
                temp_v1_data = temp_tplot['dq'][i][1][0][4]
                temp_v2_data = temp_tplot['dq'][i][1][0][6]
                temp_v3_data = temp_tplot['dq'][i][1][0][8]

                # Data array gets transposed...so we need to swap v1 and v3, while v2 stays the same
                temp_v1_data, temp_v3_data = temp_v3_data, temp_v1_data

                # Change from little endian to big endian, since pandas apparently hates little endian
                # We might want to move this into the store_data procedure eventually
                if temp_x_data.dtype.byteorder == '>':
                    temp_x_data = temp_x_data.byteswap().view(temp_x_data.dtype.newbyteorder())
                if temp_y_data.dtype.byteorder == '>':
                    temp_y_data = temp_y_data.byteswap().view(temp_y_data.dtype.newbyteorder())
                if temp_v1_data.dtype.byteorder == '>':
                    temp_v1_data = temp_v1_data.byteswap().view(temp_v1_data.dtype.newbyteorder())
                if temp_v2_data.dtype.byteorder == '>':
                    temp_v2_data = temp_v2_data.byteswap().view(temp_v2_data.dtype.newbyteorder())
                if temp_v3_data.dtype.byteorder == '>':
                    temp_v3_data = temp_v3_data.byteswap().view(temp_v3_data.dtype.newbyteorder())

                # support time-varying depends
                if len(temp_v1_data.shape) == 2:
                    temp_v1_data = np.transpose(temp_v1_data)
                if len(temp_v2_data.shape) == 2:
                    temp_v2_data = np.transpose(temp_v2_data)
                if len(temp_v3_data.shape) == 2:
                    temp_v3_data = np.transpose(temp_v3_data)

                store_data(data_name, data={'x': temp_x_data, 'y': temp_y_data, 'v1': temp_v1_data, 'v2': temp_v2_data, 'v3': temp_v3_data})
            # variable contains V1, V2 (e.g., DF as a function of energy, angle)
            elif len(temp_tplot['dq'][i][1][0]) == 8:
                temp_v1_data = temp_tplot['dq'][i][1][0][4]
                temp_v2_data = temp_tplot['dq'][i][1][0][6]

                # Data array gets transposed, so we have to swap v1 and v2
                temp_v1_data, temp_v2_data = temp_v2_data, temp_v1_data

                #Change from little endian to big endian, since pandas apparently hates little endian
                #We might want to move this into the store_data procedure eventually
                if temp_x_data.dtype.byteorder == '>':
                    temp_x_data = temp_x_data.byteswap().view(temp_x_data.dtype.newbyteorder())
                if temp_y_data.dtype.byteorder == '>':
                    temp_y_data = temp_y_data.byteswap().view(temp_y_data.dtype.newbyteorder())
                if temp_v1_data.dtype.byteorder == '>':
                    temp_v1_data = temp_v1_data.byteswap().view(temp_v1_data.dtype.newbyteorder())
                if temp_v2_data.dtype.byteorder == '>':
                    temp_v2_data = temp_v2_data.byteswap().view(temp_v2_data.dtype.newbyteorder())

                # support time-varying depends
                if len(temp_v1_data.shape) == 2:
                    temp_v1_data = np.transpose(temp_v1_data)
                if len(temp_v2_data.shape) == 2:
                    temp_v2_data = np.transpose(temp_v2_data)

                store_data(data_name, data={'x': temp_x_data, 'y': temp_y_data, 'v1': temp_v1_data, 'v2': temp_v2_data})
            # If there are 4 fields, that means it is a spectrogram
            # 6 fields is a spectrogram with a time varying Y axis
            elif len(temp_tplot['dq'][i][1][0]) == 5 or len(temp_tplot['dq'][i][1][0]) == 6:
                temp_v_data = temp_tplot['dq'][i][1][0][4]
                
                # Change from little endian to big endian, since pandas apparently hates little endian
                # We might want to move this into the store_data procedure eventually
                if temp_x_data.dtype.byteorder == '>':
                    temp_x_data = temp_x_data.byteswap().view(temp_x_data.dtype.newbyteorder())
                if temp_y_data.dtype.byteorder == '>':
                    temp_y_data = temp_y_data.byteswap().view(temp_y_data.dtype.newbyteorder())
                if temp_v_data.dtype.byteorder == '>':
                    temp_v_data = temp_v_data.byteswap().view(temp_v_data.dtype.newbyteorder())
                
                # support time-varying depends
                if len(temp_v_data.shape) == 2:
                    temp_v_data = np.transpose(temp_v_data)

                store_data(data_name, data={'x':temp_x_data, 'y':temp_y_data, 'v':temp_v_data})

            else:
                # Change from little endian to big endian, since pandas apparently hates little endian
                # We might want to move this into the store_data procedure eventually
                if temp_x_data.dtype.byteorder == '>':
                    temp_x_data = temp_x_data.byteswap().view(temp_x_data.dtype.newbyteorder())
                if temp_y_data.dtype.byteorder == '>':
                    temp_y_data = temp_y_data.byteswap().view(temp_y_data.dtype.newbyteorder())
                store_data(data_name, data={'x':temp_x_data, 'y':temp_y_data})
            
            if temp_tplot['dq'][i][3].dtype.names is not None:
                for option_name in temp_tplot['dq'][i][3].dtype.names:
                    if option_name.lower() == 'data_att':
                        arr = temp_tplot['dq'][i][3][option_name][0]
                        att_names = arr.dtype.names

                        # extract the values associated with the field names
                        att_values = arr.item()

                        # ensure the values are decoded to strings
                        att_values = [value.decode('utf-8') for value in att_values if isinstance(value, bytes)]

                        # create a dictionary with the desired mappings
                        data_att = {name.lower(): value for name, value in zip(att_names, att_values)}
                        pyspedas.tplot_tools.data_quants[data_name].attrs['data_att'] = data_att

                    if option_name.lower() not in ['color', 'colors']:
                        options(data_name, option_name, temp_tplot['dq'][i][3][option_name][0], quiet=True)

            pyspedas.tplot_tools.data_quants[data_name].attrs['plot_options']['trange'] = temp_tplot['dq'][i][4].tolist()
            pyspedas.tplot_tools.data_quants[data_name].attrs['plot_options']['create_time'] = temp_tplot['dq'][i][6]
        
            if not np.isscalar(temp_tplot['tv']): # Skip if it is a scalar
                for option_name in temp_tplot['tv'][0][0].dtype.names:
                    # the following should be set on the tplot variable, not for the entire session
                    #if option_name == 'TRANGE':
                    #    # x_range of [0, 0] causes tplot to create an empty figure
                    #    if temp_tplot['tv'][0][0][option_name][0][0] != 0 or temp_tplot['tv'][0][0][option_name][0][1] != 0:
                    #        tplot_options('x_range', temp_tplot['tv'][0][0][option_name][0])
                    if option_name == 'WSIZE':
                        tplot_options('wsize', temp_tplot['tv'][0][0][option_name][0])
                    if option_name == 'VAR_LABEL':
                        tplot_options('var_label', temp_tplot['tv'][0][0][option_name][0])
                if 'P' in temp_tplot['tv'][0][1].tolist():
                    for option_name in temp_tplot['tv'][0][1]['P'][0].dtype.names:
                        if option_name == 'TITLE':
                            tplot_options('title', temp_tplot['tv'][0][1]['P'][0][option_name][0])
 
                
            # correct legend_names array
            plt_options = pyspedas.tplot_tools.data_quants[data_name].attrs['plot_options']
            yaxis_opts = plt_options.get('yaxis_opt')
            if yaxis_opts is not None:
                yaxis_opts = plt_options.get('yaxis_opt')
                if yaxis_opts.get('legend_names') is not None:
                    lnames = pyspedas.tplot_tools.data_quants[data_name].attrs['plot_options']['yaxis_opt']['legend_names'][0]
                    if isinstance(lnames, list) or isinstance(lnames, np.ndarray):
                        pyspedas.tplot_tools.data_quants[data_name].attrs['plot_options']['yaxis_opt']['legend_names'] = [lname.decode('utf-8') for lname in lnames]
                    else:
                        pyspedas.tplot_tools.data_quants[data_name].attrs['plot_options']['yaxis_opt']['legend_names'] = [lnames.decode('utf-8')]

                # decode any other string options
                for y_key in yaxis_opts.keys():
                    if isinstance(yaxis_opts[y_key], bytes):
                        yaxis_opts[y_key] = yaxis_opts[y_key].decode("utf-8")

        #temp_tplot['tv'][0][1] is all of the "settings" variables
            #temp_tplot['tv'][0][1]['D'][0] is "device" options
            #temp_tplot['tv'][0][1]['P'][0] is "plot" options
            #temp_tplot['tv'][0][1]['X'][0] is x axis options
            #temp_tplot['tv'][0][1]['Y'][0] is y axis options
        ####################################################################
    else:
        in_file = open(filename,"rb")
        temp = pickle.load(in_file)
        num_data_quants = temp[0]
        for i in range(0, num_data_quants):
            if isinstance(temp[i+1], dict):
                # NRV variable
                pyspedas.tplot_tools.data_quants[temp[i+1]['name']] = temp[i+1]
            else:
                pyspedas.tplot_tools.data_quants[temp[i+1].name] = temp[i+1]
        pyspedas.tplot_tools.tplot_opt_glob = temp[num_data_quants+1]
        in_file.close()
    
    return



from pyspedas.tplot_tools import time_string
from pyspedas.particles.spd_units_string import spd_units_string


def slice2d_getinfo(the_slice, title=None, xtitle=None, ytitle=None, ztitle=None):
    """
    Forms various title annotations based on the slice's metadata
    """
    if title is None:
        sc = the_slice.get('spacecraft')
        if sc is None:
            sc = ''

        if not isinstance(the_slice['trange'][0], str):
            trange_str = ' -> '.join(time_string(the_slice['trange'])) + ' (' + str(the_slice['n_samples']) + ')'
        else:
            trange_str = ' -> '.join(the_slice['trange']) + ' (' + str(the_slice['n_samples']) + ')'

        title = the_slice['project_name'] + ' ' + sc + ' ' + the_slice['data_name'] + ' (' + the_slice[
            'rotation'] + ') ' + trange_str

    if xtitle is None or ytitle is None:
        xt = 'x'
        yt = 'y'

        rot = the_slice['rotation']

        if rot == 'yz':
            xt = 'y'
            yt = 'z'
        elif rot == 'xz':
            xt = 'x'
            yt = 'z'
        elif rot == 'bv':
            xt = 'B'
            yt = 'V'
        elif rot == 'be':
            xt = 'B'
            yt = 'B x V'
        elif rot == 'xvel':
            xt = 'x'
            yt = 'V'
        elif rot == 'perp':
            xt = 'V perp B'
            yt = 'B x V'
        elif rot == 'perp_xy':
            xt = '$x_{perp}$'
            yt = '$y_{perp}$'
        elif rot == 'perp_xz':
            xt = '$x_{perp}$'
            yt = '$z_{perp}$'
        elif rot == 'perp_yz':
            xt = '$y_{perp}$'
            yt = '$z_{perp}$'
        elif rot == 'b_exb':
            xt = 'B'
            yt = 'V perp B'
        elif rot == 'perp1-perp2':
            xt = 'B x V'
            yt = 'V perp B'

        # add prefix
        if the_slice['energy']:
            xyprefix = 'E'
        else:
            xyprefix = 'V'

        if the_slice['rlog']:
            xyprefix = 'log(' + xyprefix + ')'

        if xtitle is None:
            if rot not in ['perp_xy', 'perp_xz', 'perp_yz']:
                xtitle = '$' + xyprefix + '_{' + xt + '}$ (' + the_slice['xyunits'] + ')'
            else:
                xtitle = xyprefix + xt + ' (' + the_slice['xyunits'] + ')'

        if ytitle is None:
            if rot not in ['perp_xy', 'perp_xz', 'perp_yz']:
                ytitle = '$' + xyprefix + '_{' + yt + '}$ (' + the_slice['xyunits'] + ')'
            else:
                ytitle = xyprefix + yt + ' (' + the_slice['xyunits'] + ')'

        if ztitle is None:
            ztitle = spd_units_string(the_slice['units_name'])

    return {'title': title, 'xtitle': xtitle, 'ytitle': ytitle, 'ztitle': ztitle}



from pyspedas.utilities.dailynames import dailynames
from pyspedas.utilities.download import download
from pyspedas.tplot_tools import time_clip as tclip
from pyspedas.tplot_tools import cdf_to_tplot

from .config import CONFIG

def load(trange=['2006-06-01', '2006-06-02'], 
         instrument='mag',
         probe='',
         datatype='1sec',
         suffix='',
         prefix='',
         get_support_data=False, 
         varformat=None,
         varnames=[],
         downloadonly=False,
         notplot=False,
         no_update=False,
         time_clip=False):
    """
    This function loads data from the ST5 mission; this function is not meant 
    to be called directly; instead, see the wrappers:

        pyspedas.projects.st5.mag

    """

    pathformat = probe+'/'+instrument+'/%Y/st5-'+probe+'_'+datatype+'_'+instrument+'_%Y%m%d_v?.?.?.cdf'

    # find the full remote path names using the trange
    remote_names = dailynames(file_format=pathformat, trange=trange)

    out_files = []

    files = download(remote_file=remote_names, remote_path=CONFIG['remote_data_dir'], local_path=CONFIG['local_data_dir'], no_download=no_update)
    if files is not None:
        for file in files:
            out_files.append(file)

    out_files = sorted(out_files)

    if downloadonly:
        return out_files

    tvars = cdf_to_tplot(out_files, suffix=suffix, prefix=prefix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, notplot=notplot)
    
    if notplot:
        return tvars

    if time_clip:
        for new_var in tvars:
            tclip(new_var, trange[0], trange[1], suffix='')

    return tvars

    


import unittest
from pyspedas.projects.themis.common.check_args import check_args


class TestThemisCheckArg(unittest.TestCase):

    def test_basic(self):
        res_list = [['a'], ['l2']]
        self.assertEqual(check_args(probe='a', level='l2'), res_list)

    def test_multi_probes(self):
        res_list = [['a', 'b', 'c']]
        self.assertEqual(check_args(probe=' a B c   '), res_list)

    def test_multi_levels(self):
        res_list = [['l1', 'l2']]
        self.assertEqual(check_args(level='    l1 l0 l2  '), res_list)

    def test_lists(self):
        res_list = [['a', 'b', 'c'], ['l1', 'l2']]
        self.assertEqual(check_args(probe=[' a b ', 'c', 'g'], level=['l0 l1', 'l2']), res_list)

    def test_wrong_arguments(self):
        res_list = []
        self.assertEqual(check_args(probes='a b', levels='l0'), res_list)


if __name__ == '__main__':
    unittest.main()



# Copyright 2018 Regents of the University of Colorado. All Rights Reserved.
# Released under the MIT license.
# This software was developed at the University of Colorado's Laboratory for Atmospheric and Space Physics.
# Verify current version before use at: https://github.com/MAVENSDC/PyTplot

import pyspedas
from collections import OrderedDict
import logging

def tplot_rename(old_name, new_name):
    """
    This function will rename tplot variables that are already stored in memory.  
    
    Parameters
    ----------
        old_name : str 
            Old name of the Tplot Variable
        new_name : str
            New name of the Tplot Variable
         
    Returns
    -------
        None
        
    Examples
    --------
        >>> # Rename Variable 1 to Variable 2
        >>> import pyspedas
        >>> pyspedas.tplot_rename("Variable1", "Variable2")

    """

    #if old name input is a number, convert to corresponding name
    if isinstance(old_name, int):
        if isinstance(pyspedas.tplot_tools.data_quants[old_name], dict):
            old_name = pyspedas.tplot_tools.data_quants[old_name]['name']
        else:
            old_name = pyspedas.tplot_tools.data_quants[old_name].name

    # check if old name is in current dictionary
    if old_name not in pyspedas.tplot_tools.data_quants.keys():
        logging.info("The name %s is currently not in pyspedas", old_name)
        return

    # remake dictionary with new name in old name's slot

    # Why not just delete/reinsert the variable being renamed?  Doing it this way
    # preserves the ordering of variables in the dictionary.  This matches the IDL
    # behavior, where if 'tha_fit' is the first variable in the list, with index 0,
    # tplot_rename,'tha_fit', 'tha_fit_rename'
    # keeps 'tha_fit_rename' at position 0.
    # No arrays are being copied here, only references, so it's not as inefficient as
    # it might look.
    # If the variable being renamed is part of a pseudovariable, you'll end up
    # with a dangling reference to the old name.  This also matches the IDL behavior,
    # but should it?   JWL 2024/07/31

    d = pyspedas.tplot_tools.data_quants
    d2 = OrderedDict([(new_name, v) if k == old_name else (k, v) for k, v in d.items()])
    new_data_quants = d2
    for key in d2:
        if isinstance(new_data_quants[key], dict):
            # the variable is non-record varying
            new_data_quants[key]['name'] = key
        else:
            new_data_quants[key].name = key
    
    pyspedas.tplot_tools.data_quants = new_data_quants
    return



import logging
from pyspedas.tplot_tools import get_data
from pyspedas.tplot_tools import tplot_wildcard_expand


def get_any(name,tag_name):
    """

    This function returns an arbitrary value from the data_att dictionary within the metdata of a tplot variable.
    It is intended to be called by other tag-specific getters and setters.

    Parameters
    ----------
        name: str
            name of the tplot variable

        tag_name: str
            The name of the attribute to be returned

    Notes
    ----------
        The metadata values are stored in the variable's metadata at:
            metadata['data_att'][tag_name]

    Returns
    ----------
        Contents of specified tag name in the data_att structure, or None if not present.

    Examples
    ----------
        >>> import pyspedas
        >>> pyspedas.projects.themis.state(trange=['2007-06-23', '2007-06-24'], probe='a')
        >>> pyspedas.get_any('tha_pos', 'coord_sys')

    """
    metadata = get_data(name, metadata=True)
    if metadata is None:
        return None

    if metadata.get('data_att'):
        if metadata['data_att'].get(tag_name):
            return metadata['data_att'][tag_name]

    logging.debug('Tag name ' + tag_name + ' not found for variable ' + name)
    return None


def set_any(name, tag_name, tag_value):

    """

    This function sets an arbitrary attribute in the data_att dictionary within the metadata of a tplot variable.
    It is intended to be called by other tag-specific getters and setters.

    Parameters
    ----------
        name: str
            name of the tplot variable

        tag_name: str
            The name of the attribute to be stored

        tag_value: Any
            The value of the attribute to be stored

    Notes
    ----------
        The metadata values are stored in the variable's metadata at:
            metadata['data_att'][tag_name]

    Returns
    ----------
        Contents of specified tag name in the data_att structure, or None if not present.

    Examples
    ----------
        >>> import pyspedas
        >>> pyspedas.projects.themis.state(trange=['2007-06-23', '2007-06-24'], probe='a')
        >>> pyspedas.set_any('tha_pos', 'coord_sys', 'GSE')

    """
    # check that the variable exists
    data = get_data(name)
    if data is None:
        return False

    metadata = get_data(name, metadata=True)

    # note: updating the metadata dict directly updates
    # the variable's metadata in memory, so there's
    # no need to update the variable with store_data

    data_att = metadata.get('data_att')
    if data_att is None:
        metadata['data_att'] = {tag_name:tag_value}
    else:
        data_att[tag_name] = tag_value

    return True


def get_coords(name):
    """

    This function returns the coordinate system of a tplot variable

    Parameters
    ----------
        name: str
            name of the tplot variable

    Notes
    ----------
        The coordinate system is stored in the variable's metadata at:
            metadata['data_att']['coord_sys']

        See pyspedas.get_coords() to update the coordinate system

    Returns
    ----------
        Coordinate system of the tplot variable
        or
        None if the coordinate system isn't set

    Examples
    ----------
        >>> import pyspedas
        >>> pyspedas.projects.themis.state(trange=['2007-06-23', '2007-06-24'], probe='a')
        >>> pyspedas.get_coords('tha_pos')

    """

    return get_any(name,'coord_sys')

def set_coords(name, coords):
    """
    This function sets the coordinate system of a tplot variable, and updates legend names
    and axis labels if they include the coordinate system.

    Parameters
    ----------
        name: str or list[str]
            names of tplot variables to update (wildcards accepted)

        coord: str
            Abbreviated name of the coordinate system (upper case recommended)

    Notes
    ----------
        The coordinate system is stored in the variable's metadata at:
            metadata['data_att']['coord_sys']

        See pyspedas.get_coords() to return the coordinate system

    Returns
    ----------
        bool: True/False depending on if the operation was successful

    Examples
    ----------
        >>> import pyspedas
        >>> pyspedas.projects.themis.state(trange=['2007-06-23', '2007-06-24'], probe='a')
        >>> pyspedas.set_coords('tha_pos', 'GEI')
    """

    namelist = tplot_wildcard_expand(name)
    if len(namelist) == 0:
        logging.warning("set_coords: No valid tplot variables were found, returning.")
        return False

    for name in namelist:
        logging.info("Setting coordinate system for " + name)

        # Get the current value of the coordinate system, if present
        coord_in = get_coords(name)

        # Set the new coordinate system
        set_any(name,'coord_sys',coords)

        # Update other related metadata fields, if needed
        metadata=get_data(name,metadata=True)

        # should also update the legend, if it includes the coordinate system
        # for this to work, the coordinate system should be in all upper case
        if (coord_in is not None) and (metadata.get('plot_options') is not None):
            if metadata['plot_options'].get('yaxis_opt') is not None:
                if metadata['plot_options']['yaxis_opt'].get('legend_names') is not None:
                    legend = metadata['plot_options']['yaxis_opt'].get('legend_names')
                    updated_legend = [item.replace(coord_in.upper(), coords.upper()) for item in legend]
                    metadata['plot_options']['yaxis_opt']['legend_names'] = updated_legend
                if metadata['plot_options']['yaxis_opt'].get('axis_label') is not None:
                    ytitle = metadata['plot_options']['yaxis_opt'].get('axis_label')
                    metadata['plot_options']['yaxis_opt']['axis_label'] = ytitle.replace(coord_in.upper(),
                                                                                         coords.upper())
    return True

def get_units(name):
    """

    This function returns the units of a tplot variable

    Parameters
    ----------
        name: str
            name of the tplot variable

    Notes
    ----------
        The units string is stored in the variable's metadata at:
            metadata['data_att']['units']

        See set_units to update the coordinate system

    Returns
    ----------
        Units of the tplot variable
        or
        None if the coordinate system isn't set

    Examples
    ----------
        >>> import pyspedas
        >>> pyspedas.projects.themis.state(trange=['2007-06-23', '2007-06-24'], probe='a')
        >>> pyspedas.get_units('tha_pos')
    """

    return get_any(name,'units')

def set_units(name, units):
    """
    This function sets the units of a tplot variable, and updates legend names
    and axis labels if they include the units.

    Parameters
    ----------
        name: str or list[str]
            names of the tplot variables to update (wildcards accepted)

        coord: str
            Units

    Notes
    ----------
        The units are stored in the variable's metadata at:
            metadata['data_att']['units']

        See get_units to return the units

    Returns
    ----------
        bool: True/False depending on if the operation was successful

    Examples
    ----------
        >>> import pyspedas
        >>> pyspedas.projects.themis.state(trange=['2007-06-23', '2007-06-24'], probe='a')
        >>> pyspedas.set_units('tha_pos', 're')
    """

    namelist = tplot_wildcard_expand(name)
    if len(namelist) == 0:
        logging.warning("set_units: No valid tplot variables were found, returning.")
        return False

    for name in namelist:
        logging.info("Setting units for " + name)


        # Get the current units value, if present
        units_in = get_units(name)

        # Set the new value
        set_any(name,'units',units)

        # Update any other related metadata fields
        metadata = get_data(name, metadata=True)

        # should also update the legend, if it includes the coordinate system
        # for this to work, the coordinate system should be in all upper case
        if (units_in is not None) and (metadata.get('plot_options') is not None):
            if metadata['plot_options'].get('yaxis_opt') is not None:
                if metadata['plot_options']['yaxis_opt'].get('legend_names') is not None:
                    legend = metadata['plot_options']['yaxis_opt'].get('legend_names')
                    updated_legend = [item.replace(units_in, units) for item in legend]
                    metadata['plot_options']['yaxis_opt']['legend_names'] = updated_legend
                if metadata['plot_options']['yaxis_opt'].get('axis_label') is not None:
                    ytitle = metadata['plot_options']['yaxis_opt'].get('axis_label')
                    metadata['plot_options']['yaxis_opt']['axis_label'] = ytitle.replace(units_in,
                                                                                         units)

    return True



from copy import deepcopy
import numpy as np
from scipy.ndimage import shift

from pyspedas.particles.moments.moments_3d_omega_weights import moments_3d_omega_weights
from pyspedas import xyz_to_polar

# use nansum from bottleneck if it's installed, otherwise use the numpy one
try:
    import bottleneck as bn
    nansum = bn.nansum
except ImportError:
    nansum = np.nansum

def rot_mat(v1, v2):
    """
    Create a set of basis vectors based on magnetic field and velocity vectors. Used by moments_3d.

    Parameters
    ----------
    v1: ndarray
    v2: ndarray

    Returns
    -------
    ndarray
    A 3x3 rotation matrix that rotates v1 to the z' axis and v2 to the x' - z' plane
    """

    v1norm = v1/np.linalg.norm(v1)
    v2norm = v2/np.linalg.norm(v2)
    a = v1norm
    b = np.cross(a, v2)
    b = b/np.linalg.norm(b)
    c = np.cross(b,a)
    dd_out = [c, b, a]
    data_out = np.column_stack(dd_out)
    return data_out

def moments_3d(data_in, sc_pot=0, no_unit_conversion=False):
    """
    Calculates plasma moments from 3D data structure

    Parameters
    ----------
        data_in: dict
            Particle data structure with entries::

            'charge'
            'mass'
            'energy'
            'denergy'
            'theta'
            'dtheta'
            'phi'
            'dphi'
            'bins'
            'data'

        sc_pot: float
            Spacecraft potential

        no_unit_conversion: bool
            Flag indicating that datta is already in eflux and no unit
            conversion is required

    Note
    ----
        The calculations were mostly heisted from Davin Larson's IDL SPEDAS version

    Returns
    -------
    dict
        Dictionary of plasma moments with entries::

            'density'
            'flux'
            'eflux'
            'qflux'
            'mftens'
            'velocity'
            'ptens'
            'ttens'
            'vthermal'
            'avgtemp'
            'magt3'
            't3'
            'symm'
            'symm_theta'
            'symm_phi'
            'symm_ang'

    Examples
    --------

    """

    data = deepcopy(data_in)

    charge = data['charge']
    mass = data['mass']
    energy = data['energy']
    magf = data['magf']

    # Original code set the minumum energy to 0.1 eV.
    # In IDL, the energy was only set to 0.1 where e <= 0.0
    # energy[energy < 0.1] = 0.1
    energy[energy <= 0.0] = 0.1

    de = data['denergy']
    de_e = de/energy

    e_inf = energy + charge*sc_pot
    e_inf[e_inf < 0] = 0.0

    # mystery line from the IDL version
    # Less of a mystery, now that The following comments were added to the IDL version:

    # weight is a function that will decrease the integrand for electrons
    # and positive potential below a certain energy level, a gradual
    # cutoff below the pot value.This will mitigate the effect of
    # photoelectrons that may show up below the value of spacecraft
    # potential.
    # Ions should be unaffected unless the potential is negative.
    # jmm, 2025 - 11 - 24

    weight = (energy + charge*sc_pot)/de + 0.5
    weight[weight < 0] = 0
    weight[weight > 1] = 1

    domega_weight = moments_3d_omega_weights(data['theta'], data['phi'], data['dtheta'], data['dphi'])

    zero_bins = data['bins'] == 0
    if zero_bins.sum() > 0:
        data['data'][zero_bins] = 0

    data_dv = data['data']*de_e*weight*domega_weight[0, :, :]

    # density calculation
    dweight = np.sqrt(e_inf)/energy
    pardens = np.sqrt(mass/2.0)*1e-5*data_dv*dweight
    density = nansum(pardens)

    # flux calculation
    tmp = data['data']*de_e*weight*e_inf/energy
    fx = nansum(tmp*domega_weight[1, :, :])
    fy = nansum(tmp*domega_weight[2, :, :])
    fz = nansum(tmp*domega_weight[3, :, :])

    flux = np.array([fx, fy, fz])

    # velocity flux calculation
    tmp = data['data']*de_e*weight*e_inf**1.5/energy
    vfxx = nansum(tmp*domega_weight[4, :, :])
    vfyy = nansum(tmp*domega_weight[5, :, :])
    vfzz = nansum(tmp*domega_weight[6, :, :])
    vfxy = nansum(tmp*domega_weight[7, :, :])
    vfxz = nansum(tmp*domega_weight[8, :, :])
    vfyz = nansum(tmp*domega_weight[9, :, :])

    vftens = np.array([vfxx, vfyy, vfzz, vfxy, vfxz, vfyz])*np.sqrt(2.0/mass)*1e5
    mftens = vftens*mass/1e10

    # energy flux calculation (extra factor of energy)
    tmp = data['data']*de_e*weight*e_inf**2/energy
    v2f_x = nansum(tmp*domega_weight[1, :, :])
    v2f_y = nansum(tmp*domega_weight[2, :, :])
    v2f_z = nansum(tmp*domega_weight[3, :, :])
    eflux = np.array([v2f_x, v2f_y, v2f_z])

    velocity = flux/density/1e5 # km/s

    # Heat flux moments -- derived from code contributed by Terry Liu
    #
    # Terry's code has been modified to work with the units, scaling, and spacecraft potential handling in moments_3d, rather than the older
    # n_3d, j_3d, and v_3d routines.
    # JWL 2025-07-10

    eV_J = 1.602176634e-19  # conversion from eV to J

    mp=data['mass']  # mass units are eV/(km/sec)^2, for working with eflux units.  In these units, proton mass = 0.010453500
    q = eV_J

    v = np.sqrt(2.0 * energy/mp)  # convert energy array to velocity (km/sec)

    vx = v*np.cos(data['theta']/180.*np.pi)*np.cos(data['phi']/180.*np.pi)
    vy = v*np.cos(data['theta']/180.*np.pi)*np.sin(data['phi']/180.*np.pi)
    vz = v*np.sin(data['theta']/180.*np.pi)

    # Subtract bulk velocity to get thermal velocity, km/sec

    wx=vx-velocity[0]
    wy=vy-velocity[1]
    wz=vz-velocity[2]

    # thermal energy, eV
    Eth=0.5*mp*(wx**2+wy**2+wz**2)

    # Repurposed density calculation for integrating heat flux, original code made several calls to n_3d()

    data_dvx = Eth*wx*data['data'] * de_e * weight * domega_weight[0,:,:]
    data_dvy = Eth*wy*data['data'] * de_e * weight * domega_weight[0,:,:]
    data_dvz = Eth*wz*data['data'] * de_e * weight * domega_weight[0,:,:]


    dweight = np.sqrt(e_inf)/energy
    parqx = np.sqrt(mass/2.)* 1e-5 * data_dvx * dweight
    parqy = np.sqrt(mass/2.)* 1e-5 * data_dvy * dweight
    parqz = np.sqrt(mass/2.)* 1e-5 * data_dvz * dweight

    # Conversion to output units
    conv_mw = eV_J*1.0e12  # output in mW/m^2
    conv_ev = 1.0e05 # output in eV/(cm^2-sec)


    heat_x = conv_ev * nansum(parqx)
    heat_y = conv_ev * nansum(parqy)
    heat_z = conv_ev * nansum(parqz)

    qflux = [heat_x, heat_y, heat_z]


    mf3x3 = np.array([[mftens[0], mftens[3], mftens[4]], [mftens[3], mftens[1], mftens[5]], [mftens[4], mftens[5], mftens[2]]])
    pt3x3 = mf3x3 - np.outer(velocity, flux)*mass/1e5
    ptens = np.array([pt3x3[0, 0], pt3x3[1, 1], pt3x3[2, 2], pt3x3[0, 1], pt3x3[0, 2], pt3x3[1, 2]])

    t3x3 = pt3x3/density
    avgtemp = (t3x3[0, 0] + t3x3[1, 1] + t3x3[2, 2] )/3.0  # trace/3

    vthermal = np.sqrt(2.0*avgtemp/mass)

    # t3 = sp.linalg.svdvals(t3x3)

    # Calculate eigenvalues and eigenvectors
    # For some data sets, eigh() may fail to converge. Perhpas eig() is more robust?
    try:
        t3, t3evec = np.linalg.eigh(t3x3, UPLO='U')
        #t3, t3evec = np.linalg.eig(t3x3)
    except np.linalg.LinAlgError:
        # ERG can pass data arrays that are all zeros. This gives zero density and a t3x3 array full of NaNs.
        # That makes the eigenvalue calculation throw LinAlgErrors.
        # In that case, we just silently fill t3 and t3evec with NaNs and let the chips fall where they may.
        # It happens too frequently, at least for ERG HEP, to bother logging it, otherwise the logs will
        # get spammed with warnings.
        t3 = np.array([np.nan, np.nan, np.nan])
        t3evec = np.array([t3, t3, t3])

    # Note: np.linalg.eigh() returns the eigenvalues t3 in ascending order.
    # In IDL, they're not necessarily sorted.

    # Do some magical sorting and shifting
    # SPEDAS moments_3d takes magdir as a parameter, but no one seems to call it that way.
    # In that case, it defaults to [-1,1,0].

    magdir=np.array([-1.,1.,0.])
    magfn = magdir/np.linalg.norm(magdir)

    # Heuristic for identifying the t_parallel direction based on anisotropy of the eigenvalues.
    # If the mid-valued eigenvalue is less than the average of min and max, choose the index of the max,
    # otherwise the index of the min.

    s = np.argsort(t3)
    if t3[s[1]] <.5*(t3[s[0]] + t3[s[2]]):
        num=s[2]
    else:
        num=s[0]

    # Circular shift of the eigenvalue array and the columns of the eigenvector array.  This puts the
    # selected component (t_para) in component 2, and t_perp1 and t_perp2 in columns 0 and 1.
    # The order of t_perp1 and t_perp2 may differ between IDL and Python, but I am told that
    # doesn't really matter in practice.   JWL 2025-07-03

    shft = ([-1,1,0])[num]
    t3 = shift(t3,shft,mode='grid-wrap')
    t3evec = shift(t3evec,[0,shft], mode='grid-wrap')
    # This uses the magdir version of magfn, but this dot product doesn't seem to be used anywhere.
    # Instead, it is recalculated further down from magf (from the input structure) and the velocity vector
    # dot =  np.dot( magfn, t3evec[:,2] )

    bmag = np.linalg.norm(magf)
    magfn = magf/bmag

    # The next few computations are a Python rendering of the following IDL code:
    # b_dot_s = total( (magfn # [1,1,1]) * t3evec , 1)
    # dummy = max(abs(b_dot_s),num)
    #
    # rot = rot_mat(mom.magf,mom.velocity)
    # magt3x3 = invert(rot) # (t3x3 # rot)
    # mom.magt3 = magt3x3[[0,4,8]]

    # Broadcast magfn to a (3,3) matrix: each row is magfn
    magfn_matrix = np.tile(magfn, (3, 1))  # shape (3,3)

    # Element-wise multiplication
    elementwise_product = magfn_matrix * t3evec  # shape (3,3)

    # Sum across columns (axis=1)
    b_dot_s = np.sum(elementwise_product, axis=1)  # shape (3,)
    #dummy = np.max(np.abs(b_dot_s),num)

    rot = rot_mat(magf,velocity)

    # Tensor coordinate transform of t3x3
    magt3x3 = np.linalg.inv(rot) @ (t3x3 @ rot)
    magt3 = magt3x3.ravel()[[0,4,8]]
    dot = np.dot( magfn, t3evec[:,2] )
    symm_ang = np.arccos(np.abs(dot)) * 180.0/np.pi

    if dot < 0:
        t3evec = -t3evec
    symm = t3evec[:,2]

    magdir = symm

    out = xyz_to_polar(np.array([symm]))
    symm_theta = out[0,1]
    symm_phi = out[0,2]


    output = {'density': density, 
              'flux': flux,
              'eflux': eflux,
              'qflux': qflux,
              'mftens': mftens, 
              'velocity': velocity, 
              'ptens': ptens, 
              'ttens': t3x3, 
              'vthermal': vthermal,
              'avgtemp': avgtemp,
              'magt3': magt3,
              't3': t3,
              'symm': symm,
              'symm_theta': symm_theta,
              'symm_phi': symm_phi,
              'symm_ang': symm_ang,
              }
    return output



"""
Splits up 2D data into many 1D tplot variables.

Notes
-----
Similar to split_vec.pro in IDL SPEDAS.

"""

import copy
import logging
import pyspedas
from pyspedas.tplot_tools import store_data, get_data
import numpy as np
import logging

def split_vec(
        tvar,
        polar=False,
        newname=None,
        new_name=None,
        columns='all',
        suffix=None
):
    """
    Splits up 2D data into many 1D tplot variables. Takes a stored tplot vector like Vp
    and stores tplot variables Vp_x, Vp_y, Vp_z

    Parameters
    ----------
        tvar : str
            Name of tplot variable to split up
        polar : bool, optional
            If True, the input data is in polar coordinates.
            Suffix will be set to ['_mag', '_th', '_phi'].
            Default: False
        new_name : int/list, optional (Deprecated)
            The names of the new tplot variables.
            This must be the same length as the number of variables created.
        newname : int/list, optional
            The names of the new tplot variables.
            This must be the same length as the number of variables created.
            Default: None
        columns : list of ints, optional
            The specific column numbers to grab from the data.
            Default: 'all' (splits all columns)
        suffix: str
            Suffix str to be added to end of tplot variable name
            Default: None

    Returns
    -------
    list[str]
        List of variables created

    Examples
    --------
        >>> pyspedas.store_data('b', data={'x':[2,5,8,11,14,17,20], 'y':[[1,1,1,1,1,1],[2,2,5,4,1,1],[100,100,3,50,1,1],[4,4,8,58,1,1],[5,5,9,21,1,1],[6,6,2,2,1,1],[7,7,1,6,1,1]]})
        >>> pyspedas.split_vec('b')



    """
    # new_name is deprecated in favor of newname
    if new_name is not None:
        logging.info("split_vec: The new_name parameter is deprecated. Please use newname instead.")
        newname = new_name

    # Make sure the tvar is found
    if tvar not in pyspedas.tplot_tools.data_quants:
        logging.error(f"Error: {tvar} not found in memory.")
        return None

    # Give a default to the new name
    if newname is None:
        newname = tvar

    # Gather data from the tvar
    alldata = get_data(tvar)
    time = alldata[0]
    data = alldata[1]
    dim = data.shape
    metadata = get_data(tvar, metadata=True)

    has_v = False
    combined_v = None
    time_varying_v = False

    # If already size one, simply return
    if len(dim) == 1:
        return [tvar]
    # Is there depend_1 data to split?
    elif len(dim) == 2:
        if len(alldata) == 3:
            has_v = True
            combined_v = alldata[2]
            if len(combined_v.shape) == 1:
                time_varying_v = False
            else:
                time_varying_v = True

    elif len(dim) > 2:
        logging.error(f"split_vec:data array for {tvar} has {len(dim)} dimensions, too many to split")
        return None


    vec_length = dim[1]

    # Determine what the suffix list will be
    if suffix is not None:
        if vec_length > len(suffix):
            logging.error(f"split_vec error: number of columns ({vec_length}) is greater than the number of suffix entered")
    else:
        if vec_length == 3:
            if polar:
                suffix = ['_mag', '_th', '_phi']
            else:
                suffix = ["_x", "_y", "_z"]
        else:
            suffix = []
            for i in range(vec_length):
                suffix.append("_"+str(i))

    created_variables = []

    # grab column data
    if columns == 'all':
        columns = range(vec_length)

    v_idx = 0
    for i in columns:
        # if not a list
        if isinstance(i, list):
            range_start = i[0]
            range_end = i[1]
        else:
            range_start = i
            range_end = i
        split_col = list(range(range_start, range_end+1))
        split_name = newname + suffix[i]
        created_variables = created_variables + [split_name]

        if has_v and time_varying_v:
            data_for_tplot = {'x': time, 'y': data[:, split_col].squeeze(), 'v': combined_v[:,v_idx]}
        elif has_v and not time_varying_v:
            data_for_tplot = {'x': time, 'y': data[:, split_col].squeeze(), 'v': np.array([combined_v[v_idx]])}
        else:
            data_for_tplot = {'x': time, 'y': data[:, split_col].squeeze()}

        plot_options = metadata.get('plot_options', {})
        yaxis_opt = plot_options.get('yaxis_opt', {})
        labels = yaxis_opt.get('legend_names', None)
        if labels is None:
            cdf = metadata.get('CDF', {})
            labels = cdf.get('LABELS', None)

        if labels is None:
            attrs = metadata
        else:
            # Make a deep copy to avoid modifying original metadata
            attrs = copy.deepcopy(metadata)
            if 'plot_options' not in attrs:
                attrs['plot_options'] = {}
            if 'yaxis_opt' not in attrs['plot_options']:
                attrs['plot_options']['yaxis_opt'] = {}
            attrs['plot_options']['yaxis_opt']['legend_names'] = []

            if v_idx < len(labels):
                attrs['plot_options']['yaxis_opt']['legend_names'].append(labels[v_idx])

        if not store_data(split_name, data=data_for_tplot, attr_dict=attrs):
            raise Exception(f"Failed to store {split_name} in pyspedas.")
        v_idx += 1

    return created_variables



import pyspedas
# from pyspedas.projects.rbsp import load, rbsp_load_rbspice_read, rbsp_rbspice_omni, rbsp_rbspice_spin_avg
from .load import load
from .rbspice_lib.rbsp_load_rbspice_read import rbsp_load_rbspice_read
from .rbspice_lib.rbsp_rbspice_omni import rbsp_rbspice_omni
from .rbspice_lib.rbsp_rbspice_spin_avg import rbsp_rbspice_spin_avg


def rbspice(trange=['2018-11-5', '2018-11-6'],
            probe='a',
            datatype='TOFxEH',
            level='l3',
            prefix='',
            suffix='',
            force_download=False,
            get_support_data=True,
            varformat=None,
            varnames=[],
            downloadonly=False,
            notplot=False,
            no_update=False,
            time_clip=False):
    """
    This function loads data from the Radiation Belt Storm Probes Ion Composition Experiment (RBSPICE) instrument

    Parameters
    ----------
        trange : list of str, default=['2018-11-5', '2018-11-6']
            time range of interest [starttime, endtime] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        probe : str or list of str, default='a'
            Spacecraft probe name: 'a' or 'b'

        datatype: str, default='TOFxEH'
            Data type; Valid options are specific to different data levels.

        level : str, default='l3'
            Data level. Valid options: 'l1', 'l2', 'l3'

        prefix : str, optional
            The tplot variable names will be given this prefix. By default, no prefix is added.

        suffix: str, optional
            The tplot variable names will be given this suffix.  By default,
            no suffix is added.

        force_download : bool, default=False
            Download file even if local version is more recent than server version.

        get_support_data: bool, default=True
            Data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  By default, only loads in data with a
            "VAR_TYPE" attribute of "data".

        varformat: str, optional
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  By default, all variables are loaded in.

        varnames: list of str
            List of variable names to load (if not specified,
            all data variables are loaded)

        downloadonly: bool, default=False
            Set this flag to download the CDF files, but not load them into
            tplot variables

        notplot: bool, default=False
            Return the data in hash tables instead of creating tplot variables

        no_update: bool, default=False
            If set, only load data from your local cache

        time_clip: bool, default=False
            Time clip the variables to exactly the range specified in the trange keyword

    Returns
    -------
    tvars : dict or list
        List of created tplot variables or dict of data tables if notplot is True.

    Examples
    --------
    >>> rbspice_vars = pyspedas.projects.rbsp.rbspice(trange=['2018-11-5', '2018-11-6'], datatype='TOFxEH', level='l3')
    >>> tplot('rbspa_rbspice_l3_TOFxEH_proton_omni_spin')
    # Calculate the pitch angle distributions
    >>> from pyspedas.projects.rbsp.rbspice_lib.rbsp_rbspice_pad import rbsp_rbspice_pad
    >>> rbsp_rbspice_pad(probe='a', datatype='TOFxEH', level='l3')
    >>> tplot('rbspa_rbspice_l3_TOFxEH_proton_omni_0-1000keV_pad_spin')
    """

    # Valid names
    vprobe = ['a', 'b']
    vlevels = ['l1', 'l2', 'l3', 'l4']
    vdatatypesl1 = ['TOFxEH', 'TOFxEnonH', 'TOFxPHHHELT']
    vdatatypesl2 = ['TOFxEH', 'TOFxEnonH', 'TOFxPHHHELT']
    vdatatypesl3 = ['TOFxEH', 'TOFxEnonH', 'TOFxPHHHELT']
    vdatatypesl3pap = ['']  # L3PAP data is not yet supported
    vdatatypesl4 = ['']  # L4 data is not yet supported
    vdatatypes = vdatatypesl1 + vdatatypesl2 + vdatatypesl3 + vdatatypesl3pap + vdatatypesl4
    vdatatypes_lower = [vdatatype.lower() for vdatatype in vdatatypes]

    tvars = load(instrument='rbspice', trange=trange, probe=probe, datatype=datatype, level=level, prefix=prefix, suffix=suffix, force_download=force_download, get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update)

    if tvars is None or notplot or downloadonly:
        return tvars

    if not isinstance(probe, list):
        probe = [probe]

    if datatype.lower() in vdatatypes_lower:
        for prb in probe:
            # Add energy channel energy values to primary data variable,
            # create variables for individual telescopes, and set appropriate tplot options
            rbsp_load_rbspice_read(level=level, probe=prb, datatype=datatype)

            # Calculate omni-directional variable
            omni_vars = rbsp_rbspice_omni(probe=prb, datatype=datatype, level=level)
            if omni_vars:
                tvars.extend(omni_vars)

            # Calculate spin-averaged variable
            sp_avg_vars = rbsp_rbspice_spin_avg(probe=prb, datatype=datatype, level=level)
            if sp_avg_vars:
                tvars.extend(sp_avg_vars)

    return tvars



import os
import unittest
from pyspedas.tplot_tools import data_exists, del_data
import pyspedas


class LoadTestCases(unittest.TestCase):
    def test_load_mag_data(self):
        del_data()
        out_vars = pyspedas.projects.de2.mag(time_clip=True)
        self.assertTrue(data_exists("bx"))
        self.assertTrue(data_exists("by"))
        self.assertTrue(data_exists("bz"))

    def test_load_nacs_data(self):

        del_data()
        out_vars = pyspedas.projects.de2.nacs(force_download=True)
        self.assertTrue(data_exists("O_density"))
        self.assertTrue(data_exists("N_density"))

    def test_load_rpa_data(self):
        del_data()
        out_vars = pyspedas.projects.de2.rpa(prefix="de2_")
        self.assertTrue(data_exists("de2_ionDensity"))
        self.assertTrue(data_exists("de2_ionTemperature"))

    def test_load_fpi_data(self):
        del_data()
        out_vars = pyspedas.projects.de2.fpi(suffix="_de2", varnames=["TnF"])
        self.assertTrue(data_exists("TnF_de2"))

    # issue with the CDFs here, 2Nov2022
    def test_load_idm_data(self):
        del_data()
        out_vars = pyspedas.projects.de2.idm()
        self.assertTrue(data_exists("ionVelocityY"))
        self.assertTrue(data_exists("ionVelocityZ"))

    def test_load_wats_data(self):
        del_data()
        out_vars = pyspedas.projects.de2.wats()
        self.assertTrue(data_exists("density"))
        self.assertTrue(data_exists("Tn"))

    def test_load_vefi_data(self):
        del_data()
        out_vars = pyspedas.projects.de2.vefi(trange=["1981-12-06", "1981-12-07"])
        self.assertTrue(data_exists("spectA"))
        self.assertTrue(data_exists("spectB"))
        self.assertTrue(data_exists("spectC"))

    def test_load_lang_data(self):
        del_data()
        out_vars = pyspedas.projects.de2.lang(time_clip=True)
        self.assertTrue(data_exists("plasmaDensity"))
        self.assertTrue(data_exists("electronTemp"))

    def test_load_notplot(self):
        del_data()
        out_vars = pyspedas.projects.de2.mag(notplot=True)
        self.assertTrue("bz" in out_vars)

    def test_downloadonly(self):
        del_data()
        files = pyspedas.projects.de2.mag(downloadonly=True, trange=["1983-2-16", "1983-2-17"])
        self.assertTrue(os.path.exists(files[0]))


if __name__ == "__main__":
    unittest.main()



from .maven_load import load_data
import pyspedas.projects.maven.spdf as spdf_load

# This routine was originally in maven/__init__.py, until being moved to its own file.
# Please refer to __init__.py if you need to see the revision history before it was moved.

maven_load = load_data

def sta(
    trange=["2016-01-01", "2016-01-02"],
    level="l2",
    datatype=None,
    suffix="",
    prefix="",
    varformat=None,
    get_support_data=False,
    auto_yes=True,
    downloadonly=False,
    varnames=[],
    spdf=False,
):
    """
    Function to load MAVEN STA data.

    Parameters
    ----------
    trange : list, optional
        Time range of the data in the format ["start_date", "end_date"].
        Defaults to ["2016-01-01", "2016-01-02"].
    level : str, optional
        Data level. Defaults to "l2".
    datatype : str, optional
        Data type. Defaults to "c0-64e2m" (spdf) or None (all data loaded) (MAVEN SDC)
        Valid options (for MAVEN SDC)::

            2a, c0, c2, c4, c6, c8, ca, cc, cd, ce, cf, d0, d1, d4, d6, d7, d8, d9, da, db

    varformat : str, optional
        Variable format. Defaults to None.
    suffix: str
        The tplot variable names will be given this suffix.
        Default: '', no suffix is added.
    prefix: str
        The tplot variable names will be given this prefix.
        Default: '', no prefix is added.
    get_support_data : bool, optional
        Whether to retrieve support data. Defaults to False.
    auto_yes : bool, optional
        Whether to automatically answer "yes" to prompts. Defaults to True.
    downloadonly : bool, optional
        Whether to only download the data without loading it. Defaults to False.
    varnames : list, optional
        List of variable names to load. Defaults to [].
    spdf : bool, optional
        Whether to use the SPDF library for loading data. Defaults to False.

    Returns
    -------
    dict
        Dictionary of loaded data variables.
    """

    if spdf:
        if datatype is None:
            datatype = "c0-64e2m"
        return spdf_load.static(
            trange=trange,
            level=level,
            datatype=datatype,
            varformat=varformat,
            get_support_data=get_support_data,
            downloadonly=downloadonly,
            varnames=varnames,
        )
    if datatype is None:
        datatype = None
    return maven_load(
        instruments="sta",
        start_date=trange[0],
        end_date=trange[1],
        type=datatype,
        level=level,
        varformat=varformat,
        suffix=suffix,
        prefix=prefix,
        get_metadata=True,
        get_support_data=get_support_data,
        auto_yes=auto_yes,
        download_only=downloadonly,
        varnames=varnames,
    )



"""Test functions in the utilites folder."""

import os
import unittest
from copy import deepcopy
from pyspedas.projects import themis
from pyspedas import (
    del_data,
    tplot,
    tplot_options,
    tplot_names,
    store_data,
    wildcard_expand,
    tname_byindex,
    tplot_wildcard_expand,
    tindex_byname,
    data_exists,
    tnames,
)
from pyspedas.utilities.config_testing import TESTING_CONFIG

# Whether to display plots during testing
global_display = TESTING_CONFIG["global_display"]
# Directory to save testing output files
output_dir = TESTING_CONFIG["local_testing_dir"]
# Ensure output directory exists
save_dir = os.path.join(output_dir, "utilities")
if not os.path.exists(save_dir):
    os.makedirs(save_dir)

tp_data = {"x": [1, 2, 3], "y": [4, 5, 6]}


class UtilTestCases(unittest.TestCase):
    """Test functions related to tplot variable indexing and wildcards."""

    def test_wildcard_expand(self):
        # Test wildcard expansion against defined list
        master_list = ["Foo", "foo", "tha_pos_gsm", "thb_pos_gsm", "mms1_mec_r_sm"]
        # Literal match
        result = wildcard_expand(master_list, "Foo")
        self.assertEqual(result, ["Foo"])
        # Results should be ordered following patterns, not master list
        result = wildcard_expand(master_list, ["foo", "Foo"])
        self.assertEqual(result, ["foo", "Foo"])
        # Wildcard expansion
        result = wildcard_expand(master_list, "th?_*_*")
        self.assertEqual(result, ["tha_pos_gsm", "thb_pos_gsm"])
        # Case sensitive (by default), no matches found
        result = wildcard_expand(master_list, "*SM")
        self.assertEqual(result, [])
        # Case insensitive
        result = wildcard_expand(master_list, "*SM", case_sensitive=False)
        self.assertEqual(result, ["tha_pos_gsm", "thb_pos_gsm", "mms1_mec_r_sm"])
        # Null pattern
        result = wildcard_expand(master_list, None)
        self.assertEqual(result, [])
        # Master list and pattern both null
        result = wildcard_expand(None, None)
        self.assertEqual(result, [])
        # Master list not a list (Formerly an error, now accepted)
        result = wildcard_expand("foo", "foo")
        self.assertEqual(result, ["foo"])
        # Master list as space delimited string  (used to be an error, now accepted)
        master_list_str = "Foo foo tha_pos_gsm thb_pos_gsm mms1_mec_r_sm"
        result = wildcard_expand(master_list_str, "foo")
        self.assertEqual(result, ["foo"])
        # Master list and search list as space delimited strings
        result = wildcard_expand(master_list_str, "foo bar")
        self.assertEqual(result, ["foo"])
        # Search list is a list containing some space delimited strings
        result = wildcard_expand(master_list_str, ["Foo", "foo", "thb_pos_gsm mms1_mec_r_sm"])
        self.assertEqual(result, ["Foo", "foo", "thb_pos_gsm", "mms1_mec_r_sm"])
        # Master list is empty
        result = wildcard_expand([], "foo")
        self.assertEqual(result, [])
        result = wildcard_expand([""], "foo")
        self.assertEqual(result, [])
        # Patterns are empty
        result = wildcard_expand(master_list_str, [])
        self.assertEqual(result, [])
        result = wildcard_expand(master_list_str, "")
        self.assertEqual(result, [])
        # Do not split on spaces
        result = wildcard_expand(["foo bar", "baz quux"], "foo bar", split_whitespace=False)
        self.assertEqual(result, ["foo bar"])

    def test_wildcard_quiet(self):
        master = ["string1", "string2", "string3"]
        patt1 = "notfound"
        with self.assertLogs(level="WARNING") as log:
            # This should produce a warning
            result = wildcard_expand(master, patt1)
            self.assertTrue("No match found" in log.output[0])
            # With quiet==True, there should be no warning
            result = wildcard_expand(master, patt1, quiet=True)
            self.assertIsNotNone(result)
            self.assertTrue(len(log.output) == 1)

    def test_tplot_wildcard_quiet(self):
        del_data("*")
        store_data("foo", deepcopy(tp_data))
        patt1 = "notfound"
        with self.assertLogs(level="WARNING") as log:
            # This should produce a warning
            result = tplot_wildcard_expand(patt1)
            self.assertTrue("No match found" in log.output[0])
            self.assertFalse(result)
            # With quiet==True, there should be no warning
            result = tplot_wildcard_expand(patt1, quiet=True)
            self.assertTrue(len(log.output) == 1)
            del_data("*")
            # Should warn about empty list of tplot variable names
            result = tplot_wildcard_expand(patt1)
            self.assertTrue("is empty" in log.output[1])
            # len(log.output) is now 2
            # With quiet==True, there should be no warning
            result = tplot_wildcard_expand(patt1, quiet=True)
            self.assertTrue(len(log.output) == 2)

    def test_tplot_wildcard_expand(self):
        # Test wildcard expansion against current tplot variables
        del_data("*")
        master_list = ["Foo", "foo", "tha_pos_gsm", "thb_pos_gsm", "mms1_mec_r_sm"]
        for var in master_list:
            store_data(var, deepcopy(tp_data))
        # Add a composite variable
        store_data("composite", data=["foo", "tha_pos_gsm", "thb_pos_gsm"])
        tplot_names()
        tn = tplot_names(quiet=True)
        print(tn)
        # Literal match
        result = tplot_wildcard_expand("Foo")
        self.assertEqual(result, ["Foo"])
        # Results should be ordered following patterns, not master list
        result = tplot_wildcard_expand(["foo", "Foo"])
        self.assertEqual(result, ["foo", "Foo"])
        # Wildcard expansion
        result = tplot_wildcard_expand("th?_*_*")
        self.assertEqual(result, ["tha_pos_gsm", "thb_pos_gsm"])
        # Composite variable
        result = tplot_wildcard_expand("composite")
        self.assertEqual(result, ["composite"])
        # Case sensitive (by default), no matches found
        result = tplot_wildcard_expand("*SM")
        self.assertEqual(result, [])
        # Case insensitive
        result = tplot_wildcard_expand("*SM", case_sensitive=False)
        self.assertEqual(result, ["tha_pos_gsm", "thb_pos_gsm", "mms1_mec_r_sm"])
        # Mixed strings and integers with repeated values
        result = tplot_wildcard_expand(["Foo", 1, 2, 2, "tha*"], case_sensitive=False)
        self.assertEqual(result, ["Foo", "foo", "tha_pos_gsm"])
        # Null pattern
        result = tplot_wildcard_expand(None)
        self.assertEqual(result, [])

    def test_tindex_byname(self):
        # Test lookups of tplot variable indices by name
        del_data("*")
        store_data("foo", data=deepcopy(tp_data))
        store_data("bar", data=deepcopy(tp_data))
        result = tindex_byname("foo")
        self.assertEqual(result, 0)
        result = tindex_byname("bar")
        self.assertEqual(result, 1)
        result = tindex_byname("doesnt_exist")
        self.assertEqual(result, None)

    def test_tname_byindex(self):
        # Test lookups of tplot variable names by indices
        del_data("*")
        # Out of bounds (no tplot variables yet)
        result = tname_byindex(0)
        self.assertEqual(result, None)
        store_data("foo", data=deepcopy(tp_data))
        store_data("bar", data=deepcopy(tp_data))
        result = tname_byindex(0)
        self.assertEqual(result, "foo")
        result = tname_byindex(1)
        self.assertEqual(result, "bar")
        # Index too large
        result = tname_byindex(2)
        self.assertEqual(result, None)
        # Negative index
        result = tname_byindex(-1)
        self.assertEqual(result, None)

    def test_plotting_wildcards_indices(self):
        # Test tplot using wildcards
        del_data("*")
        themis.state(probe=["a", "b", "c", "d", "e"])
        # Wildcard pattern
        tplot_options("title", "Positions of THEMIS A B C D E")
        tplot("th?_pos", display=global_display, save_png=os.path.join(save_dir, "tplot_wildcard.png"))
        # List of integer indices
        indices = [tindex_byname("tha_pos"), tindex_byname("thb_pos"), tindex_byname("thc_pos")]
        tplot_options("title", "Positions of THEMIS A B C")
        tplot(indices, display=global_display, save_png=os.path.join(save_dir, "tplot_byindex.png"))
        # List including a bad name
        tplot_options("title", "Positions of THEMIS A B C D E")
        tplot(["th?_pos", "doesnt_exist"], display=global_display, save_png=os.path.join(save_dir, "tplot_list_onebad.png"))

    def test_tplot_names_embedded_spaces(self):
        # Tplot names can have embedded spaces, make sure they're treated correctly
        del_data("*")
        t = [1.0, 2.0, 3.0]
        d = [1, 2, 3]
        store_data("embedded spaces", data={"x": t, "y": d})
        names = tnames()
        self.assertEqual(names, ["embedded spaces"])
        names = tplot_names()
        self.assertEqual(names, ["embedded spaces"])
        del_data("embedded spaces")
        self.assertFalse(data_exists("embedded spaces"))
        store_data("embedded spaces", data={"x": t, "y": d})
        del_data("*")
        self.assertFalse(data_exists("embedded spaces"))
        store_data("embedded spaces", data={"x": t, "y": d})
        # This should warn about 'd', 'e', and 'f' not found, but not warn about 'embedded' or 'spaces'
        vals = tplot_wildcard_expand(["embedded spaces", "d e f"])
        self.assertEqual(vals, ["embedded spaces"])


if __name__ == "__main__":
    unittest.main()



"""
Python implementation of IDL roundsig function.

Rounds numbers to a specified number of significant figures.
"""

import numpy as np


def roundsig(x, sigfig=None, uncertainty=None):
    """
    Round numbers to a specified number of significant figures.

    Python translation of IDL roundsig function.

    Parameters
    ----------
    x : array_like
        Input values to round.
    sigfig : int, optional
        Number of significant figures (default=1).
    uncertainty : float, optional
        If provided, rounds to nearest multiple of uncertainty.

    Returns
    -------
    ndarray
        Rounded values.
    """
    x = np.asarray(x, dtype=float)

    # Handle uncertainty case (IDL keyword_set(unc))
    if uncertainty is not None:
        return uncertainty * np.round(x / uncertainty)

    # Default sigfig to 1 if not provided
    if sigfig is None:
        sigfig = 1

    # Work with copy to avoid modifying input
    lx = x.copy()

    # Track negative values
    neg = lx < 0

    # Track zero values
    wz = lx == 0
    if np.any(wz):
        lx[wz] = 1.0

    # Take log10 of absolute values
    lx = np.log10(np.abs(lx))

    # Calculate exponent and fractional parts
    e = np.floor(lx - sigfig)
    f = lx - e

    # Calculate mantissa
    man = np.round(10**f)

    # Reconstruct the rounded value
    rx = man * (10.0**e)

    # Restore negative signs
    if np.any(neg):
        rx[neg] = -rx[neg]

    # Restore zeros
    if np.any(wz):
        rx[wz] = 0

    return rx



import os
import unittest
from pyspedas.tplot_tools import data_exists
from pyspedas.tplot_tools import get_data
import pyspedas
from pyspedas.projects.poes.load import load


class LoadTestCases(unittest.TestCase):
    def test_load_notplot(self):
        sem_vars = pyspedas.projects.poes.sem(notplot=True)
        self.assertTrue('ted_ele_tel0_low_eflux' in sem_vars)

    def test_load_sem_data(self):
        sem_vars = pyspedas.projects.poes.sem(time_clip=True)
        self.assertTrue(data_exists('ted_ele_tel0_low_eflux'))

    def test_load_probe_regression(self):
        sem_vars = pyspedas.projects.poes.sem(probe='metop1', time_clip=True)
        self.assertTrue(data_exists('ted_ele_tel0_low_eflux'))
        m = get_data('ted_ele_tel0_low_eflux', metadata=True)
        # Workaround for cdflib globalattsget bug
        md_sn = m['CDF']['GATT']['Source_name']
        if isinstance(md_sn,list):
            md_sn = md_sn[0]
        self.assertTrue(md_sn == 'MetOp1')
        sem_vars = pyspedas.projects.poes.sem(probe='metop2', time_clip=True)
        m = get_data('ted_ele_tel0_low_eflux', metadata=True)
        # workaround for cdflib globalattsget bug
        md_sn = m['CDF']['GATT']['Source_name']
        if isinstance(md_sn,list):
            md_sn = md_sn[0]
        self.assertTrue(md_sn == 'MetOp2')

    def test_downloadonly(self):
        files = pyspedas.projects.poes.sem(downloadonly=True, probe='noaa19')
        self.assertTrue(os.path.exists(files[0]))

    def test_ncei_server(self):
        vars = load(trange=['1999-01-03', '1999-01-04'], probe=['noaa15'], ncei_server=True, time_clip=True)
        self.assertTrue('geogLL' in vars)


if __name__ == '__main__':
    unittest.main()



import logging
from pyspedas.tplot_tools import get_data
import numpy as np

from pyspedas.projects.elfin.load import load
from pyspedas.projects.elfin.epd.postprocessing import epd_l1_postprocessing, epd_l2_postprocessing

def epd_load(trange=['2022-08-19', '2022-08-19'],
        probe='a',
        datatype='pef',
        level='l1',
        type_='nflux',
        get_support_data=False,
        varformat=None,
        varnames=[],
        downloadonly=False,
        notplot=False,
        no_update=False,
        time_clip=True,
        nspinsinsum=None,
        fullspin=False,
        PAspec_energies=None,
        PAspec_energybins=None,
        Espec_LCfatol=None,
        Espec_LCfptol=None,
        force_download=False):
    """
    This function loads data from the ELFIN Energetic Particle Detector (EPD) and process L1 and L2 data.

    Parameters for Load Routine
    ---------------------------
        trange : list of str
            Time range of interest [starttime, endtime]. Format can be
            ['YYYY-MM-DD','YYYY-MM-DD'] or ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']
            Default: ['2022-08-19', '2022-08-19']

        probe: str, optional
            Spacecraft identifier. Options are 'a' and 'b'.
            Default: 'a'

        level: str, optional.
            Data level. Options are 'l1' and 'l2'
            Default: 'l1'

        get_support_data: bool, optional
            If True, data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.
            Default: only loads in data with a "VAR_TYPE" attribute of "data".

        varformat: str, optional
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.
            Default: all variables are loaded in.

        varnames: list of str, optional
            List of variable names to load.
            Default: all data variables are loaded.

        downloadonly: bool, optional
            If True, only downloads the CDF files without loading them into tplot variables. 
            Default: False.

        notplot: bool, optional
            If True, returns data in hash tables instead of creating tplot variables. 
            Default: False.

        no_update: bool
            If True, loads data only from the local cache.
            Default: False.

        time_clip: bool
            If True, clips the variables to the exact range specified in the trange. 
            Default: True.

        force_download: bool
            Download file even if local version is more recent than server version
            Default: False


    Parameters for L1 data
    ----------------------
        datatype: str, optional. 
            Data type of L1 data. Options are 'pef' , 'pif', 'pes', 'pis'
            Default: 'pef'

        type: str, optional
            Calibrated data type of L1 data. Options are 'raw', 'cps', 'nflux', 'eflux'
            Default: 'nflux'

        nspinsinsum: int, optional
            Number of spins in sum which is needed by the L1 calibration function. Options are 16 or 32.
            Default: 16
    
    Parameters for L2 data
    ----------------------
        fullspin: bool, optional.
            If True, generate L2 full spin spectrogram.
            Default: L2 half spin spectrogram is generated.

        PAspec_energybins: list of tuple of int, optional
            Specified the energy bins used for generating L2 pitch angle spectrogram.
            If both 'PAspec_energybins' and 'PAspec_energies' are set, 'energybins' takes precedence.
            Default: [(0,2),(3,5), (6,8), (9,15)].
            
        PAspec_energies: list of tuple of float, optional
            Specifies the energy range for each bin in the L2 pitch angle spectrogram.
            Example: energies=[(50.,160.),(160.,345.),(345.,900.),(900.,7000.)]
            If both 'energybins' and 'energies' are set, 'energybins' takes precedence.
            Default:
            Energy and energybin table::

                channel     energy_range    energy_midbin
                0           50-80           63.2
                1           80-120          97.9
                2           120-160         138.5
                3           160-210         183.3
                4           210-270         238.1
                5           270-345         305.2
                6           345-430         385.1
                7           430-630         520.4
                8           630-900         752.9
                9           900-1300        1081.6
                10          1300-1800       1529.7
                11          1800-2500       2121.3
                12          2500-3350       2893.9
                13          3350-4150       3728.6
                14          4150-5800       4906.1
                15          5800+           6500.0

        Espec_LCfatol: float, optional
            Tolerance angle for para and anti flux in generating L2 energy spectrogram. 
            A positive value makes the loss cone/antiloss cone smaller by this amount. 
            Default: 22.25 deg.

        Espec_LCfptol: float, optional
            Tolerance angle for perp flux in generating L2 energy spectrogram. 
            A negative value means a wider angle for perp flux.
            Default: -11 deg.

    Returns
    -------
        list of str
            List of tplot variables created.

    Examples
    --------

        >>> import pyspedas
        >>> from pyspedas import tplot
        >>> elf_vars = pyspedas.projects.elfin.epd(probe='b', trange=['2021-01-01', '2021-01-02'], datatype='pif')
        >>> tplot(['elb_pif_nflux', 'elb_pif_spinper'])

        >>> import pyspedas
        >>> from pyspedas import tplot
        >>> elf_vars = pyspedas.projects.elfin.epd(probe='a', trange=['2022-08-19', '2022-08-19'], level='l2')
        >>> tplot(['ela_pef_fs_Epat_nflux', 'ela_pef_hs_Epat_nflux'. 'ela_pef_pa', 'ela_pef_tspin'])

    """
    logging.info("ELFIN EPD: START LOADING.")
    
    tvars = load(instrument='epd', probe=probe, trange=trange, level=level, datatype=datatype,
                 get_support_data=get_support_data, varformat=varformat, varnames=varnames, downloadonly=downloadonly,
                 notplot=notplot, time_clip=time_clip, no_update=no_update, force_download=force_download,)

    logging.info("ELFIN EPD: LOADING END.")
    if notplot or downloadonly:
        return tvars
    elif not tvars:
        logging.error('ELFIN EPD: cannot load data.')
        return tvars
   
    CALIBRATED_TYPE_UNITS = {
        "raw": "counts/sector",
        "cps": "counts/s",
        "nflux": "#/(s-cm$^2$-str-MeV)",
        "eflux": "keV/(s-cm$^2$-str-MeV)",
    }

    if type_ in ("cal", "calibrated") or type_ not in CALIBRATED_TYPE_UNITS.keys():
        type_ = "nflux"
        
    if level == "l1":
        l1_tvars = epd_l1_postprocessing(tvars, trange=trange, type_=type_, nspinsinsum=nspinsinsum,
                                     unit=CALIBRATED_TYPE_UNITS[type_])
        return l1_tvars

    elif level == "l2":
        logging.info("ELFIN EPD L2: START PROCESSING.")
        # check whether input type is allowed
        if type_ not in ("nflux","eflux"):
            logging.warning(f"fluxtype {type_} is not allowed in l2 data, change to nflux!")
            type_ = "nflux"

        res = 'hs' if fullspin is False else 'fs'

        # if 32 sector data is needed, pass the variables with 32
        tvars_32 = [tvar for tvar in tvars if '_32' in tvar]
        tvars_16 = [tvar.replace('_32', '') for tvar in tvars_32]
        tvars_other = list(set(tvars) - set(tvars_32) - set(tvars_16))
        
        tvars_input = tvars_other + tvars_16
        if len(tvars_32) != 0 :
            data = get_data(tvars_32[0])
            if np.any(~np.isnan(data.y)) : # if any 32 sector data is not nan
                tvars_input = tvars_other + tvars_32
        
        l2_tvars = epd_l2_postprocessing(
            tvars_input,
            fluxtype=type_,
            res=res,
            PAspec_energies=PAspec_energies,
            PAspec_energybins=PAspec_energybins,
            Espec_LCfatol=Espec_LCfatol,
            Espec_LCfptol=Espec_LCfptol,)
        
        return l2_tvars
    else:
        raise ValueError(f"Unknown level: {level}")

    return tvars



import os

CONFIG = {
    "remote_data_dir_ae": "https://wdc.kugi.kyoto-u.ac.jp/ae_provisional/",
    "remote_data_dir_dst": "http://wdc.kugi.kyoto-u.ac.jp/",
    "local_data_dir": "pydata/geom_indices/kyoto/",
}

# override local data directory with environment variables
if os.environ.get("SPEDAS_DATA_DIR"):
    CONFIG["local_data_dir"] = os.sep.join(
        [os.environ["SPEDAS_DATA_DIR"], "geom_indices/kyoto/"]
    )



import logging
import numpy as np
from pyspedas.utilities.dailynames import dailynames
from pyspedas.utilities.download import download
from pyspedas.tplot_tools import time_clip as tclip
from pyspedas.tplot_tools import cdf_to_tplot, get_data, store_data, options

from .config import CONFIG


def load(
    trange=["2019-02-01", "2019-02-02"],
    site=None,
    suffix="",
    get_support_data=False,
    varformat=None,
    varnames=[],
    downloadonly=False,
    notplot=False,
    no_update=False,
    time_clip=False,
):
    """
    Loads data from the Magnetic Induction Coil Array (MICA)

    Parameters
    ----------
    trange : list of str, optional
        Time range of interest with the format ['YYYY-MM-DD','YYYY-MM-DD'] or 
        to specify more or less than a day ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']. 
        Defaults to ['2019-02-01', '2019-02-02'].
    site : str, optional
        The site to load data for.
    suffix : str, optional
        The tplot variable names will be given this suffix. Defaults to ''.
    get_support_data : bool, optional
        Data with an attribute "VAR_TYPE" with a value of "support_data" will be loaded into tplot. 
        By default, only loads in data with a "VAR_TYPE" attribute of "data". Defaults to False.
    varformat : str, optional
        The file variable formats to load into tplot. Wildcard character "*" is accepted. 
        By default, all variables are loaded in. Defaults to None.
    varnames : list of str, optional
        List of variable names to load. If not specified, all data variables are loaded. Defaults to [].
    downloadonly : bool, optional
        Set this flag to download the CDF files, but not load them into tplot variables. Defaults to False.
    notplot : bool, optional
        Return the data in hash tables instead of creating tplot variables. Defaults to False.
    no_update : bool, optional
        If set, only load data from your local cache. Defaults to False.
    time_clip : bool, optional
        Time clip the variables to exactly the range specified in the trange keyword. Defaults to False.

    Returns
    -------
    list
        List of tplot variables created.
        If downloadonly is set to True, the list of files downloaded will be returned instead.
    """

    if site is None:
        logging.error("A valid MICA site code name must be entered.")
        logging.error("Current site codes include: ")
        logging.error(
            "NAL, LYR, LOR, ISR, SDY, IQA, SNK, MCM, SPA, JBS, NEV, HAL, PG2[3,4,5]"
        )
        return

    pathformat = site.upper() + "/%Y/%m/mica_ulf_" + site.lower() + "_%Y%m%d_v??.cdf"

    # find the full remote path names using the trange
    remote_names = dailynames(file_format=pathformat, trange=trange)

    out_files = []

    files = download(
        remote_file=remote_names,
        remote_path=CONFIG["remote_data_dir"],
        local_path=CONFIG["local_data_dir"],
        no_download=no_update,
    )
    if files is not None:
        for file in files:
            out_files.append(file)

    out_files = sorted(out_files)

    if downloadonly:
        return out_files

    tvars = cdf_to_tplot(
        out_files,
        suffix="_" + site.upper() + suffix,
        get_support_data=get_support_data,
        varformat=varformat,
        varnames=varnames,
        notplot=notplot,
    )

    # remove values > 1000; taken from IDL SPEDAS version
    for out_var in tvars:
        if out_var[0:7] == "spectra":
            times, data, freq = get_data(out_var)
            w_fill = np.where(data > 1000.0)
            data[w_fill] = np.nan
            store_data(out_var, data={"x": times, "y": data, "v": freq})
            options(out_var, "spec", True)
            options(out_var, "Colormap", "spedas")
            options(out_var, "zlog", False)

    if notplot:
        return tvars

    if time_clip:
        for new_var in tvars:
            tclip(new_var, trange[0], trange[1], suffix="")

    return tvars




import logging
import pyspedas
from pyspedas.tplot_tools import tnames, tplot_copy
import numpy


def subtract_average(
        names,
        newname=None,
        suffix=None,
        overwrite=None,
        median=None
):
    """
    Subtracts the average or median from data.

    Parameters
    ----------
    names: str/list of str
        List of tplot variable names to process (wildcards accepted)
    newname: str/list of str, optional
        List of new names for tplot variables.
        Default: None. If not given, then a suffix is applied.
    suffix: str, optional
        A suffix to apply.
        Default: '-d'.
    overwrite: bool, optional
        If set, then tplot variables are replaced.
        Default: None
    median: float, optional
        If it is 0 or not set, then it computes the mean.
        Otherwise, it computes the median.
        Default: None.

    Returns
    -------
    list of str
        List of new tplot variables created

    Examples
    --------
        >>> pyspedas.store_data('a', data={'x':[0,4,8,12,16], 'y':[1,2,3,4,5]})
        >>> pyspedas.subtract_average('a')
        >>> pyspedas.tplot(['a','a-d'])

    """

    old_names = tnames(names)

    if len(old_names) == 0:
        return
    elif len(old_names) < 1:
        logging.error('Subtract Average error: No tplot names were provided.')
        return

    if suffix is None:
        if median:
            suffix = '-m'
        else:
            suffix = '-d'

    if overwrite is not None:
        n_names = old_names
    elif newname is None:
        n_names = [s + suffix for s in old_names]
    else:
        n_names = newname

    if isinstance(n_names, str):
        n_names = [n_names]

    if len(n_names) != len(old_names):
        n_names = [s + suffix for s in old_names]

    old_names = tnames(names)

    for old_idx, old in enumerate(old_names):
        new = n_names[old_idx]

        if new != old:
            tplot_copy(old, new)

        data = pyspedas.tplot_tools.data_quants[new].values
        # Subtracting the average will fail if data is not a floating point type
        if data.dtype.kind != 'f':
            data=numpy.float64(data)
        dim = data.shape
        if median:
            if len(dim) == 1:
                if not numpy.isnan(data).all():
                        data -= numpy.nanmedian(data, axis=0)
            else:
                for i in range(dim[1]):
                    if not numpy.isnan(data[:, i]).all():
                        data[:, i] -= numpy.nanmedian(data[:, i], axis=0)
            ptype = 'Median'
        else:
            if len(dim) == 1:
                if not numpy.isnan(data).all():
                    data -= numpy.nanmean(data, axis=0)
            else:
                for i in range(dim[1]):
                    if not numpy.isnan(data[:, i]).all():
                        data[:, i] -= numpy.nanmean(data[:, i], axis=0)
            ptype = 'Mean'

        pyspedas.tplot_tools.data_quants[new].values = data

        logging.info('Subtract ' + ptype + ' was applied to: ' + new)

    return n_names



"""
This is a python version of the wavetest.pro script in IDL SPEDAS.
It is an example program for wavelet analysis using NINO3 SST dataset.
This script assumes that the data file sst_nino3.dat is in the same directory as the script.

Reference: Torrence, C. and G. P. Compo, 1998: A Practical Guide to
           Wavelet Analysis. <I>Bull. Amer. Meteor. Soc.</I>, 79, 61-78.

See also: wavelet98.py
"""

import os
import numpy as np
import matplotlib.pyplot as plt
from pyspedas import wavelet98, wave_signif
from pyspedas.utilities.config_testing import TESTING_CONFIG, test_data_download_file

# Whether to display plots during testing
global_display = TESTING_CONFIG["global_display"]
# Directory to save testing output files
output_dir = TESTING_CONFIG["local_testing_dir"]
# Ensure output directory exists
analysis_dir = "analysis_tools"
save_dir = os.path.join(output_dir, analysis_dir)
if not os.path.exists(save_dir):
    os.makedirs(save_dir)
# Directory with IDL SPEDAS validation files
validation_dir = TESTING_CONFIG["remote_validation_dir"]


def wavetest(noplot=False):
    """
    Python translation of wavetest.pro
    Wavelet analysis of NINO3 SST dataset.
    """

    # Download tplot files
    # The SPEDAS script that creates the file: general/tools/python_validate/wavelet_python_validate.pro

    data_file = test_data_download_file(
        validation_dir, analysis_dir, "sst_nino3.dat", save_dir
    )

    # Check if data file exists
    if not os.path.exists(data_file):
        raise FileNotFoundError(f"Data file not found: {data_file}")

    # Load SST time series data
    sst = np.loadtxt(data_file)
    n = len(sst)

    # Normalize by standard deviation (not necessary)
    sst = sst - np.mean(sst)

    # Parameters
    # dt: Time step (quarterly data)
    # xrange: Plotting range
    # pad: Padding for wavelet transform
    # s0: Start scale (3 months)
    # dj: Sub-octaves per octave
    # mother: Mother wavelet type
    dt = 0.25  # quarterly data
    time = np.arange(n) * dt + 1871.0
    xrange = [1870, 2000]

    # Wavelet parameters
    pad = 1
    s0 = dt  # start at scale of 3 months
    dj = 0.25  # 4 sub-octaves per octave
    j1 = 9.0 / dj  # this says do 9 powers-of-two with dj sub-octaves each
    mother = "MORLET"

    # Perform wavelet transform
    print("Computing wavelet transform...")
    wave, scale, period, signif, daughter_ret, sst_recon, fft_theor_out = wavelet98(
        sst, dt, pad=pad, dj=dj, j_scales=j1, s0=s0, mother=mother, recon=True
    )
    power = np.abs(wave) ** 2  # Wavelet power spectrum

    # Global wavelet spectrum (GWS) [Eqn(10)]
    global_ws = np.mean(power, axis=0)

    # Compute significance levels for the global wavelet spectrum
    # This corresponds to the IDL code:
    # global_signif = WAVE_SIGNIF(sst, dt, scale, 1, LAG1=0.0, DOF=dof, MOTHER=mother, CDELTA=Cdelta, PSI0=psi0)
    dof = len(sst) - scale  # Degrees of freedom
    global_signif, wave_outputs = wave_signif(
        sst, dt, scale, 1, lag1=0.0, dof=dof, mother=mother
    )
    Cdelta = wave_outputs["Cdelta"]

    # Add cone of influence (COI) for the wavelet power spectrum
    # This corresponds to the IDL code:
    # x = [time[0], time, MAX(time)]
    # y = [MAX(period), coi, MAX(period)]
    # POLYFILL, x, y, ORIEN=+45, SPACING=0.5, COLOR=color, NOCLIP=0, THICK=1
    coi = period / np.sqrt(2)  # Calculate COI based on wavelet period
    # Interpolate COI to match the length of the time array
    coi_interp = np.interp(time, time[: len(coi)], coi)
    if not noplot:
        plt.fill_between(
            time, coi_interp, np.max(period), color="white", alpha=0.7, hatch="///"
        )

    # Reconstruction variance (Parseval's theorem) [Eqn(14)]
    # Adjust scale shape for broadcasting
    recon_variance = (
        dj * dt / (np.sum(scale) * len(sst)) * np.sum(power / scale[None, :])
    )
    print(f"Reconstructed variance = {recon_variance:.4f} degC^2")

    # Additional reconstruction analysis if sst_recon is available
    if sst_recon is not None and len(sst_recon) > 1:
        # Calculate variance of reconstructed time series
        recon_variance_actual = np.var(
            sst_recon, ddof=0
        )  # Use population variance like IDL MOMENT

        # RMS of Reconstruction [Eqn(11)]
        rms_error = np.sqrt(np.sum((sst - sst_recon) ** 2) / n)

        # Original variance
        original_variance = np.var(
            sst, ddof=0
        )  # Use population variance like IDL MOMENT

        print()
        print("        ******** RECONSTRUCTION ********")
        print(f"original variance = {original_variance:.6f} degC^2")
        print(f"reconstructed var = {recon_variance_actual:.6f} degC^2")
        print(f"Ratio = {recon_variance_actual/original_variance:.6f}")
        print(f"root-mean-square error of reconstructed sst = {rms_error:.6f} degC")
        print()
        if mother.upper() == "DOG":
            print("Note: for better reconstruction with the DOG, you need")
            print("      to use a very small s0.")
        print()

    # RMS of Reconstruction [Eqn(11)] - original calculation
    rms_error = np.sqrt(np.mean((sst - np.mean(sst)) ** 2))
    print(
        f"RMS error of reconstruction = {rms_error:.4f}"
    )  # Keep original for compatibility

    # Compute scale-average between 2-8 years [Eqn(24)]
    # Following IDL implementation more closely
    avg_idx = np.where((period >= 2) & (period < 8))[0]
    if len(avg_idx) > 0:
        # Normalize power by scale (like IDL power_norm = power/scale_avg)
        scale_expanded = scale[None, :]  # Expand for broadcasting
        power_norm = power / scale_expanded

        # Scale-average calculation [Eqn(24)] - matches IDL exactly
        scale_avg = dj * dt / Cdelta * np.sum(power_norm[:, avg_idx], axis=1)
    else:
        scale_avg = np.zeros(len(time))

    # Compute scale-average significance levels
    # This corresponds to the IDL code:
    # scaleavg_signif = WAVE_SIGNIF(sst, dt, scale, 2, GWS=global_ws, SIGLVL=0.90, DOF=[2,7.9], MOTHER=mother)
    scaleavg_signif, _ = wave_signif(
        sst, dt, scale, 2, gws=global_ws, siglvl=0.90, dof=[2, 7.9], mother=mother
    )

    # Ensure scaleavg_signif matches the length of time
    if np.isscalar(scaleavg_signif):
        scaleavg_signif = np.full_like(time, scaleavg_signif, dtype=np.float64)
    elif len(scaleavg_signif) == 1:
        scaleavg_signif = np.full_like(time, scaleavg_signif[0], dtype=np.float64)

    # Plotting
    if not noplot:
        create_wavelet_plot(
            time,
            sst,
            time,
            period,
            power,
            global_ws,
            scale_avg,
            xrange,
            global_signif,
            scaleavg_signif,
        )

    print("\nWavelet analysis completed!")
    return {
        "time": time,
        "sst": sst,
        "period": period,
        "power": power,
        "global_ws": global_ws,
        "scale_avg": scale_avg,
    }


def create_wavelet_plot(
    time,
    sst,
    time_wv,
    period,
    power,
    global_ws,
    scale_avg,
    xrange,
    global_signif,
    scaleavg_signif,
):
    """
    Create the wavelet analysis plot similar to IDL version
    """

    fig = plt.figure(figsize=(12, 10))

    # Plot 1: Time series (top panel, full width)
    _ = plt.subplot2grid((4, 4), (0, 0), colspan=3)
    plt.plot(time, sst, "b-", linewidth=1)
    plt.xlim(xrange)
    plt.xlabel("Time (year)")
    plt.ylabel("NINO3 SST (°C)")
    plt.title("a) NINO3 Sea Surface Temperature (seasonal)")
    plt.grid(True, alpha=0.3)

    # Adjust subplot positioning to give more space for labels and prevent overlap
    plt.subplots_adjust(
        left=0.12, bottom=0.08, right=0.95, top=0.95, hspace=0.4, wspace=0.3
    )

    # Plot 2: Wavelet power spectrum (middle left, 3/4 width)
    _ = plt.subplot2grid((4, 4), (1, 0), colspan=3, rowspan=2)

    # Create contour plot with IDL-like styling
    levels = [0.5, 1, 2, 4]  # Contour levels for power spectrum
    colors = ["lightblue", "yellow", "orange", "red"]

    X, Y = np.meshgrid(time_wv, period)
    _ = plt.contourf(X, Y, power.T, levels=levels, colors=colors, extend="max")

    # Cone of influence (placeholder)
    coi = np.max(period) * np.ones_like(time_wv)  # Placeholder for COI calculation
    plt.fill_between(
        time_wv, coi, np.max(period), color="white", alpha=0.7, hatch="///"
    )

    plt.xlim(xrange)
    plt.ylim([128, 0.5])  # Extended range like requested (128 at top, 0.5 at bottom)
    plt.yscale("log")
    # Create powers-of-2 ticks like IDL
    yticks = [0.5, 1, 2, 4, 8, 16, 32, 64, 128]
    plt.yticks(yticks, [str(int(y)) if y >= 1 else str(y) for y in yticks])
    plt.xlabel("Time (year)")
    plt.ylabel("Period (years)")
    plt.title("b) Wavelet Power Spectrum (contours at 0.5,1,2,4°C²)")

    # Plot 3: Global wavelet spectrum (middle right, 1/4 width)
    ax3 = plt.subplot2grid((4, 4), (1, 3), rowspan=2)
    plt.plot(global_ws, period, "b-", linewidth=2)
    plt.plot(global_signif, period, "r--", linewidth=1)
    # Linear scales but inverted like plot 2
    plt.xlim([0, 4])  # Linear x-axis from 0 to 4
    plt.ylim([128, 0.5])  # Same inverted range as plot 2 but linear
    plt.yscale("log")
    ax3.set_yscale("log")  # Explicitly set y-axis to linear using axes object
    # Use same tick positions as plot 2 but no labels
    yticks = [0.5, 1, 2, 4, 8, 16, 32, 64, 128]
    plt.yticks(yticks, [""] * len(yticks))
    plt.xlabel("Power (°C²)")
    plt.title("c) Global")
    plt.text(0.7, 0.8, "95%", transform=ax3.transAxes)
    plt.grid(True, alpha=0.3)

    # Plot 4: Scale-average time series (bottom panel, 3/4 width)
    _ = plt.subplot2grid((4, 4), (3, 0), colspan=3)
    plt.plot(time_wv, scale_avg, "b-", linewidth=2)
    # Plot significance as horizontal line
    if np.isscalar(scaleavg_signif):
        plt.axhline(y=scaleavg_signif, color="r", linestyle="--", linewidth=1)
    else:
        plt.plot(time_wv, scaleavg_signif, "r--", linewidth=1)
    plt.xlim(xrange)
    plt.ylim([0, np.max(scale_avg) * 1.25])  # Auto-scale like IDL
    plt.xlabel("Time (year)")
    plt.ylabel("Avg variance (°C²)")
    plt.title("d) 2-8 yr Scale-average Time Series")
    plt.grid(True, alpha=0.3)

    # Save figure to output directory and then show it (optionally)
    output_path = os.path.join(save_dir, "wavelet_analysis_nino3_sst.png")
    fig.savefig(output_path, dpi=300, facecolor="white")
    if global_display:
        plt.show()


if __name__ == "__main__":
    # Run the wavelet test
    result = wavetest()



# Copyright 2018 Regents of the University of Colorado. All Rights Reserved.
# Released under the MIT license.
# This software was developed at the University of Colorado's Laboratory for Atmospheric and Space Physics.
# Verify current version before use at: https://github.com/MAVENSDC/PyTplot

from collections import OrderedDict

import os
import sys
import xarray as xr

# xarray options
xr.set_options(keep_attrs=True)


# Global Variables
data_quants = OrderedDict()
tplot_opt_glob = dict(
    tools="xpan,crosshair,reset",
    min_border_top=12,
    min_border_bottom=12,
    title_align="center",
    window_size=[800, 800],
    title_size="12",
    title_text="",
    crosshair=True,
    data_gap=0,
    black_background=False,
    axis_font_size=12,
    axis_tick_num=[
        (0, 1000000000),
        (3, 1),
    ],
    y_axis_zoom=False,
)
lim_info = {}
extra_layouts = {}

from .get_y_range import get_y_range
from .tplot_rename import tplot_rename
from .del_data import del_data
from .replace_metadata import replace_metadata
from .store_data import store_data, store
from .get_data import get_data, get
from .str_to_float_fuzzy import str_to_float_fuzzy
from .tplot_names import tplot_names
from .wildcard_routines import (
    wildcard_expand,
    tplot_wildcard_expand,
    tname_byindex,
    tindex_byname,
)
from .options import options
from .xlim import xlim
from .ylim import ylim
from .zlim import zlim
from .tlimit import tlimit
from . import spedas_colorbar
from .exporters.tplot_save import tplot_save
from .tnames import tnames
from .convert_tplotxarray_to_pandas_dataframe import (
    convert_tplotxarray_to_pandas_dataframe,
)
from .importers.tplot_restore import tplot_restore
from .is_pseudovariable import is_pseudovariable
from .count_traces import count_traces
from .get_timespan import get_timespan
from .tplot_options import tplot_options
from .tplot_copy import tplot_copy
from .replace_data import replace_data
from .replace_metadata import replace_metadata
from .get_ylimits import get_ylimits
from .rgb_color import rgb_color
from .timebar import timebar
from .databar import databar
from .timespan import timespan
from .timestamp import timestamp
from .time_double import time_float, time_double, time_float_one
from .time_string import time_string, time_datetime, time_string_one
from .importers.sts_to_tplot import sts_to_tplot
from .data_att_getters_setters import set_coords, get_coords, set_units, get_units
from .data_exists import data_exists
from .link import link
from .tres import tres

from .tplot_math.tinterp import tinterp
from .tplot_math.add_across import add_across
from .tplot_math.add import add
from .tplot_math.avg_res_data import avg_res_data
from .tplot_math.clip import clip
from .tplot_math.crop import crop
from .tplot_math.deflag import deflag
from .tplot_math.degap import degap
from .tplot_math.derive import derive
from .tplot_math.divide import divide
from .tplot_math.interp_nan import interp_nan
from .tplot_math.tinterp import tinterp
from .tplot_math.join_vec import join_vec
from .tplot_math.multiply import multiply
from .tplot_math.spec_mult import spec_mult
from .tplot_math.split_vec import split_vec
from .tplot_math.subtract import subtract
from .tplot_math.pwr_spec import pwr_spec
from .tplot_math.tsmooth import smooth
from .tplot_math.tsmooth import tsmooth
from .tplot_math.tdotp import tdotp
from .tplot_math.tcrossp import tcrossp
from .tplot_math.tnormalize import tnormalize
from .tplot_math.subtract_average import subtract_average
from .tplot_math.subtract_median import subtract_median
from .tplot_math.clean_spikes import clean_spikes
from .tplot_math.time_clip import time_clip
from .tplot_math.tkm2re import tkm2re
from .tplot_math.makegap import makegap
from .tplot_math.tdeflag import tdeflag
from .tplot_math.pwrspc import pwrspc
from .tplot_math.tpwrspc import tpwrspc
from .tplot_math.dpwrspc import dpwrspc
from .tplot_math.tdpwrspc import tdpwrspc

from .reduce_spec_dataset import reduce_spec_dataset
from .MPLPlotter.lineplot import lineplot
from .MPLPlotter.specplot import specplot, specplot_make_1d_ybins
from .MPLPlotter.get_var_label_ticks import get_var_label_ticks
from .MPLPlotter.var_labels import var_label_panel
from .MPLPlotter.ctime import ctime
from .MPLPlotter.highlight import highlight
from .MPLPlotter.annotate import annotate
from .MPLPlotter.tplot import tplot
from .importers.cdf_to_tplot import cdf_to_tplot
from .importers.netcdf_to_tplot import netcdf_to_tplot
from .exporters.tplot_ascii import tplot_ascii



import numpy as np
from pyspedas import tnames
from pyspedas.tplot_tools import time_float
from pyspedas.tplot_tools import get_data, options, store_data, ylim, zlim

from ..load import load
from ..get_gatt_ror import get_gatt_ror


from typing import List, Optional

def pwe_efd(
    trange: List[str] = ['2017-04-01', '2017-04-02'],
    datatype: str = 'E_spin',
    level: str = 'l2',
    suffix: str = '',
    coord: str = 'dsi',
    get_support_data: bool = False,
    varformat: Optional[str] = None,
    varnames: List[str] = [],
    downloadonly: bool = False,
    notplot: bool = False,
    no_update: bool = False,
    uname: Optional[str] = None,
    passwd: Optional[str] = None,
    time_clip: bool = False,
    ror: bool = True,
    force_download: bool = False,
) -> List[str]:
    """
    This function loads data from the PWE experiment from the Arase mission

    Parameters
    ----------
        trange : list of str
            time range of interest [starttime, endtime] with the format
            'YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']
            Default: ['2017-04-01', '2017-04-02']

        datatype: str
            Data type; Valid options: 'E256Hz', 'E64Hz', 'E_spin', 'pot', 'pot8Hz', 'spec'
            Default: 'E_spin'

        level: str
            Data level; Valid options: 'l2'
            Default: 'l2'

        coord: str
            Coordinate system to use. Valid options: 'dsi', 'wpt'
            Default: 'dsi'

        suffix: str
            The tplot variable names will be given this suffix.  Default: None

        get_support_data: bool
            If True, data with an attribute "VAR_TYPE" with a value of "support_data"
            will be loaded into tplot.  Default: False

        varformat: str
            The file variable formats to load into tplot.  Wildcard character
            "*" is accepted.  Default: None (all variables loaded)

        varnames: list of str
            List of variable names to load. If list is empty or not specified,
            all data variables are loaded. Default: [] (all variables loaded)

        downloadonly: bool
            Set this flag to download the CDF files, but not load them into
            tplot variables. Default: False

        notplot: bool
            Return the data in hash tables instead of creating tplot variables
            Default: False

        no_update: bool
            If set, only load data from your local cache
            Default: FAlse

        time_clip: bool
            Time clip the variables to exactly the range specified in the trange keyword
            Default: False

        ror: bool
            If set, print PI info and rules of the road

        uname: str
            User name. Default: None

        passwd: str
            Password. Default: None

        force_download: bool
            Download file even if local version is more recent than server version
            Default: False

    Returns
    -------
        List of tplot variables created.

    Examples
    --------
    >>> import pyspedas
    >>> from pyspedas import tplot
    >>> pwe_efd_vars = pyspedas.projects.erg.pwe_efd(trange=['2017-03-27', '2017-03-28'])
    >>> tplot('erg_pwe_efd_l2_E_spin_Eu_dsi')

    """
    initial_notplot_flag = False
    if notplot:
        initial_notplot_flag = True
    file_res = 3600. * 24
    prefix = 'erg_pwe_efd_'+level+'_'

    if ('64' in datatype) or ('256' in datatype):
        if '64' in datatype:
            mode = '64Hz'
        elif '256' in datatype:
            mode = '256Hz'
        md = 'E'+mode
        pathformat = 'satellite/erg/pwe/efd/'+level+'/'+md + \
            '/%Y/%m/erg_pwe_efd_'+level+'_'+md+'_'+coord+'_%Y%m%d_v??_??.cdf'
        prefix += md + '_' + coord + '_'
        if coord == 'wpt':
            component = ['Eu_waveform', 'Ev_waveform']
        elif coord == 'dsi':
            component = ['Ex_waveform', 'Ey_waveform',
                         'Eu_offset'+'_'+mode, 'Ev_offset'+'_'+mode]

    else:
        pathformat = 'satellite/erg/pwe/efd/'+level+'/'+datatype + \
            '/%Y/%m/erg_pwe_efd_'+level+'_'+datatype+'_%Y%m%d_v??_??.cdf'
        prefix += datatype + '_'
        if 'spin' in datatype:
            component = ['Eu', 'Ev', 'Eu1', 'Ev1', 'Eu2', 'Ev2']
            labels = ['Ex', 'Ey']
        if datatype == 'pot':
            component = ['Vu1', 'Vu2', 'Vv1', 'Vv2']
        if datatype == 'pot8Hz':
            component = ['Vu1_waveform_8Hz', 'Vu2_waveform_8Hz',
                         'Vv1_waveform_8Hz', 'Vv2_waveform_8Hz']

    loaded_data = load(pathformat=pathformat, trange=trange, level=level, datatype=datatype, file_res=file_res, prefix=prefix, suffix=suffix, get_support_data=get_support_data,
                       varformat=varformat, varnames=varnames, downloadonly=downloadonly, notplot=notplot, time_clip=time_clip, no_update=no_update, uname=uname, passwd=passwd, force_download=force_download)

    if (len(loaded_data) > 0) and ror:

        try:
            gatt = get_gatt_ror(downloadonly, loaded_data)

            # --- print PI info and rules of the road

            print(' ')
            print(' ')
            print(
                '**************************************************************************')
            print(gatt["LOGICAL_SOURCE_DESCRIPTION"])
            print('')
            print('Information about ERG PWE EFD')
            print('')
            print('PI: ', gatt['PI_NAME'])
            print("Affiliation: ",gatt["PI_AFFILIATION"])
            print('')
            print('RoR of ERG project common: https://ergsc.isee.nagoya-u.ac.jp/data_info/rules_of_the_road.shtml.en')
            print(
                'RoR of PWE/EFD: https://ergsc.isee.nagoya-u.ac.jp/mw/index.php/ErgSat/Pwe/Efd')
            print('')
            print('Contact: erg_pwe_info at isee.nagoya-u.ac.jp')
            print(
                '**************************************************************************')
        except:
            print('printing PI info and rules of the road was failed')

    if initial_notplot_flag or downloadonly:
        return loaded_data

    time_min_max = time_float(trange)
    if 'spin' in datatype:
        for elem in component:
            t_plot_name = prefix + elem + '_dsi'
            options(t_plot_name, 'ytitle', elem + ' vector in DSI')
            options(t_plot_name, 'legend_names', labels)
            # ylim settings because pyspedas.tplot_tools.timespan() doesn't affect in ylim.
            # May be it will be no need in future.
            get_data_vars = get_data(t_plot_name)
            if get_data_vars[0][0] < time_min_max[0]:
                min_time_index = np.where(
                    (get_data_vars[0] <= time_min_max[0]))[0][-1]
            else:
                min_time_index = 0
            if time_min_max[1] < get_data_vars[0][-1]:
                max_time_index = np.where(
                    time_min_max[1] <= get_data_vars[0])[0][0]
            else:
                max_time_index = -1
            ylim_min = np.nanmin(
                get_data_vars[1][min_time_index:max_time_index])
            ylim_max = np.nanmax(
                get_data_vars[1][min_time_index:max_time_index])
            ylim(t_plot_name, ylim_min, ylim_max)

    if ('64' in datatype) or ('256' in datatype) or (datatype == 'pot8Hz'):
        for elem in component:
            t_plot_name = prefix+elem
            get_data_vars = get_data(t_plot_name)
            dl_in = get_data(t_plot_name, metadata=True)
            time1 = get_data_vars[0]
            data = np.where(get_data_vars[1] <= -
                            1e+30, np.nan, get_data_vars[1])
            dt = get_data_vars[2]
            ndt = dt.size
            ndata = data.size
            time_new = (np.tile(time1, (ndt, 1)).T + dt * 1e-3).reshape(ndata)
            data_new = data.reshape(ndata)
            store_data(t_plot_name, data={
                       'x': time_new, 'y': data_new}, attr_dict=dl_in)
            options(t_plot_name, 'ytitle', '\n'.join(t_plot_name.split('_')))
            # ylim settings because pyspedas.tplot_tools.timespan() doesn't affect in ylim.
            # May be it will be no need in future.
            if time_new[0] < time_min_max[0]:
                min_time_index = np.where((time_new <= time_min_max[0]))[0][-1]
            else:
                min_time_index = 0
            if time_min_max[1] < time_new[-1]:
                max_time_index = np.where(time_min_max[1] <= time_new)[0][0]
            else:
                max_time_index = -1
            ylim_min = np.nanmin(data_new[min_time_index:max_time_index])
            ylim_max = np.nanmax(data_new[min_time_index:max_time_index])
            ylim(t_plot_name, ylim_min, ylim_max)

    if datatype == 'pot':
        for elem in component:
            t_plot_name = prefix+elem
            options(t_plot_name, 'ytitle', elem+' potential')

    if datatype == 'spec':
        options(tnames(prefix + '*spectra*'), 'spec', 1)
        options(tnames(prefix + '*spectra*'), 'colormap', 'jet')
        options(tnames(prefix + '*spectra*'), 'zlog', 1)
        ylim(prefix + 'spectra', 0, 100)
        zlim(prefix + 'spectra', 1e-6, 1e-2)
        options(tnames(prefix + '*spectra*'), 'ysubtitle', '[Hz]')
        options(tnames(prefix + '*spectra*'), 'ztitle', '[mV^2/m^2/Hz]')
        for t_plot_name in (tnames(prefix + '*spectra*')):
            options(t_plot_name, 'ytitle', '\n'.join(t_plot_name.split('_')))

    return loaded_data



import pyspedas
from pyspedas.cotrans_tools.minvar import minvar
from .rotmat_set_coords import rotmat_set_coords
from pyspedas.tplot_tools import get_data, store_data, time_double, time_string
import numpy as np


def minvar_matrix_make(in_var_name,
                       tstart=None,
                       tstop=None,
                       twindow=None,
                       tslide=None,
                       newname=None,
                       evname=None,
                       tminname=None,
                       tmidname=None,
                       tmaxname=None):
    """
    Use an input variable to make rotation matrices to rotate vectors into a minimum variance coordinate system.

    Parameters
    ----------
    in_var_name : str
        Name of the input variable that defines the coordinate system
    tstart: str
        Start time of the interval for which MVA matrices will be created.
    tend: str
        End time of the interval for which MVA matrices will be created.
    twindow: float
        Duration (in seconds) of the window used to produce each matrix. Defaults to
        entire time series.
    tslide: float
        Offset time (in seconds) between each successive window start time. Defaults to twindow/2.
        Set to 0 to produce only a single matrix.
    newname: str
        Name of the tplot variable to receive the MVA rotation matrices.
    evname: str
        Name of the tplot variable to receive the eigenvalues at each window.
    tminname: str
         Name of the tplot variable to receive the basis vectors for the minimum variance direction.
    tmidname: str
         Name of the tplot variable to receive the basis vectors for the intermediate variance direction.
    tmaxname: str
         Name of the tplot variable to receive the basis vectors for the maximum variance direction.

    Returns
    -------
    list of str
        The list of tplot variables produced.

    """
    data = get_data(in_var_name)
    input_coords = pyspedas.get_coords(in_var_name)

    if tstart is None:
        start_d = data.times[0]
    else:
        start_d = time_double(tstart)

    if tstop is None:
        stop_d = data.times[-1]
    else:
        stop_d = time_double(tstop)

    if (twindow is None) or (twindow > (stop_d - start_d)):
        twindow = stop_d - start_d

    # sets it to something large enough that it will only generate one
    # matrix in the default case
    if (tslide is None) or (tslide <= 0) :
        tslide = twindow / 2.0

    if newname is None:
        newname = in_var_name + '_mva_mat'

    current_d = start_d

    # Exact number of windows: always at least one, plus however many tslides fit in the space after the first window
    o_num = 1 + int((stop_d - start_d - twindow)/tslide)

    o_times = np.zeros(o_num)
    o_lams = np.zeros((o_num, 3))
    o_eigs = np.zeros((o_num, 3, 3))

    i = 0

    while current_d + twindow <= stop_d:
        if i >= o_num:
            break
        # output time for the mva matrix is the midpoint time for the interval
        o_times[i] = current_d + twindow / 2.0

        idx = np.argwhere((data.times >= current_d) & (data.times <= (current_d + twindow))).flatten()

        if len(idx) == 0:
            current_d += tslide
            i += 1
            continue

        in_data = data.y[idx, :]

        try:
            out = minvar(in_data)
        except np.linalg.LinAlgError:
            current_d += tslide
            i += 1
            continue

        if out is not None:
            v_rot, o_eig, o_lam = out
            o_lams[i, :] = o_lam
            o_eigs[i, :, :] = o_eig.transpose()

        i += 1
        current_d += tslide

        if tslide == 0:
            break

    o_d = {'x': o_times, 'y': o_eigs}

    out_vars = []

    saved = store_data(newname, data=o_d)

    if saved:
        # Set input and output coordinates for this rotation matrix
        rotmat_set_coords(newname, input_coords, 'MVA')
        out_vars.append(newname)

    if evname is not None:
        store_data(evname, data={'x': o_times, 'y': o_lams})
        out_vars.append(evname)

    if tminname is not None:
        store_data(tminname, data={'x': o_times, 'y': o_eigs[:, 2, :]})
        out_vars.append(tminname)

    if tmidname is not None:
        store_data(tmidname, data={'x': o_times, 'y': o_eigs[:, 1, :]})
        out_vars.append(tmidname)

    if tmaxname is not None:
        store_data(tmaxname, data={'x': o_times, 'y': o_eigs[:, 0, :]})
        out_vars.append(tmaxname)

    return out_vars



import logging

from pyspedas.tplot_tools import get, store, del_data, tnames, tplot_rename, options, tplot
from pyspedas.tplot_tools import time_clip as tclip

from pyspedas.projects.elfin.epd.calibration_l2 import epd_l2_Espectra, epd_l2_PAspectra
from pyspedas.projects.elfin.epd.calibration_l1 import calibrate_epd

def epd_l1_postprocessing(
    tplotnames,
    trange=None,
    type_=None,
    nspinsinsum=None,
    unit=None,
):
    """
    Calibrates data from the Energetic Particle Detector (EPD) and sets dlimits.

    Parameters
    ----------
        tplotnames : list of str
            The tplot names of EPD data to be postprocessed.

        trange : list of str
            Time range of interest [starttime, endtime] with the format
            ['YYYY-MM-DD','YYYY-MM-DD'] or to specify more or less than a day
            ['YYYY-MM-DD/hh:mm:ss','YYYY-MM-DD/hh:mm:ss']

        type_ : str, optional
            Desired data type. Options: 'raw', 'cps', 'nflux', 'eflux'.

        nspinsinsum : int, optional
            Number of spins in sum which is needed by the calibration function.

        unit : str, optional
            Units of the data.


    Returns
    ----------
        List of tplot variables created.

    """

    tplotnames = tplotnames.copy()

    # Calibrate spinper to turn it to seconds
    FGM_SAMPLE_RATE = 80.0

    for name in filter(lambda n: "spinper" in n, tplotnames):
        d = get(name)
        cal_spinper = d.y / FGM_SAMPLE_RATE
        store(name, {"x": d.times, "y": cal_spinper}, metadata=get(name, metadata=True))

    if nspinsinsum is None:
        tn = tnames("*nspinsinsum*")
        if tn:
            nspin = get(tn[0])
            nspinsinsum = nspin.y if nspin is not None else 1
        else:
            nspinsinsum = 1

    new_tvars = []

    for name in tplotnames:
        if "energies" in name:
            del_data(name)
            continue
        if "sectnum" in name:
            options(name, 'ytitle', name)
            new_tvars.append(name)
            continue
        if "spinper" in name:
            options(name, 'ytitle', name)
            options(name, 'ysubtitle','[sec]')
            new_tvars.append(name)
            continue
        if "nspinsinsum" in name:
            options(name, 'ytitle', name)
            new_tvars.append(name)
            continue
        if "nsectors" in name:
            options(name, 'ytitle', name)
            new_tvars.append(name)
            continue
        
        new_name = f"{name}_{type_}"
        tplot_rename(name, new_name)
        options(new_name, 'ytitle', new_name) # set units for elx_pef variable
        options(new_name, 'ysubtitle', unit)
        new_tvars.append(new_name)

        calibrate_epd(new_name, trange=trange, type_=type_, nspinsinsum=nspinsinsum)

    # TODO: Set units and tplot options (obey no_spec)

    return new_tvars


def epd_l2_postprocessing(
    tplotnames,
    fluxtype='nflux',
    res='hs',
    datatype='e',
    PAspec_energies = None,
    PAspec_energybins = None,
    Espec_LCfatol = None,
    Espec_LCfptol = None,
):
    """
    Process ELF EPD L2 data and generate omni, para, anti, perp flux spectra.

    Parameters
    ----------
        tplotnames : list of str
            The tplot names of EPD data to be postprocessed.

        fluxtype: str, optional
            Type of flux spectra. 
            Options: 'nflux' for number flux, 'eflux' for energy flux.
            Default is 'nflux'.

        res: str, optional
            Resolution of spectra. 
            Options: 'hs' for half spin, 'fs' for full spin. 
            Default is 'hs'.

        datatype: str, optional
            Type of data. 
            Options: 'e' for electron data, 'i' for ion data
            Default is 'e'.
        
        PAspec_energybins: list of tuple of int, optional
            Specified the energy bins used for generating l2 pitch angle spectra. 
            Default is [(0,2),(3,5), (6,8), (9,15)]. If both 'PAspec_energybins' and 'PAspec_energies' 
            are set, 'energybins' takes precedence
            
        PAspec_energies: list of tuple of float, optional
            Specifies the energy range for each bin in the l2 pitch angle spectra.
            Example: energies=[(50.,160.),(160.,345.),(345.,900.),(900.,7000.)]
            If both 'energybins' and 'energies' are set, 'energybins' takes precedence.
            Energy and energybin table:
            channel     energy_range    energy_midbin
            0           50-80           63.2
            1           80-120          97.9
            2           120-160         138.5
            3           160-210         183.3
            4           210-270         238.1
            5           270-345         305.2
            6           345-430         385.1
            7           430-630         520.4
            8           630-900         752.9
            9           900-1300        1081.6
            10          1300-1800       1529.7
            11          1800-2500       2121.3
            12          2500-3350       2893.9
            13          3350-4150       3728.6
            14          4150-5800       4906.1
            15          5800+           6500.0

        Espec_LCfatol: float, optional
            Tolerance angle for para and anti flux. A positive value makes the loss 
            cone/antiloss cone smaller by this amount. 
            Default is 22.25 deg.

        Espec_LCfptol: float, optional
            Tolerance angle for perp flux. A negative value means a wider angle for 
            perp flux.
            Default is -11 deg.

    Returns
    ----------
        List of tplot variables created.
    """

    tplotnames = tplotnames.copy()

    flux_tname = [name for name in tplotnames if f"p{datatype}f_{res}_Epat_{fluxtype}" in name]
    LC_tname = [name for name in tplotnames if f"_{res}_LCdeg" in name]
    
    if len(flux_tname) != 1:
        logging.error(f'{len(flux_tname)} flux tplot variables is found!')
        return
    
    if len(LC_tname) != 1:
        logging.error(f'{len(LC_tname)} LC tplot variables is found!')
        return
    
    logging.info("ELFIN EPD L2: START ENERGY SPECTOGRAM.")
    # get energy spectra in four directions
    E_tvar = epd_l2_Espectra(flux_tname[0], LC_tname[0], LCfatol=Espec_LCfatol, LCfptol=Espec_LCfptol)

    logging.info("ELFIN EPD L2: START PITCH ANGLE SPECTOGRAM.")
    # get pitch angle spectra
    #PA_tvar = epd_l2_PAspectra(flux_tname[0], energies=[(60, 200),(300, 1000)])
    PA_tvar = epd_l2_PAspectra(flux_tname[0], energies=PAspec_energies, energybins=PAspec_energybins)

    return E_tvar + PA_tvar